#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import logging

from pyworkflow.utils import LoggingConfigurator

import os
import sys
import time
import argparse

from pyworkflow.protocol import (getProtocolFromDb,
                                 STATUS_FINISHED, STATUS_ABORTED, STATUS_FAILED,
                                 STATUS_RUNNING, STATUS_SCHEDULED, STATUS_SAVED,
                                 STATUS_LAUNCHED, Set, Protocol, MAX_SLEEP_TIME)
from pyworkflow.constants import PROTOCOL_UPDATED

stopStatuses = [STATUS_FINISHED, STATUS_ABORTED, STATUS_FAILED]


class RunScheduler:
    """ Check that all dependencies are met before launching a run. """

    def __init__(self, args, logger):
        self._args = args
        # Enter to the project directory and load protocol from db
        self.protocol = self._loadProtocol()
        self.project = self.protocol.getProject()
        self.log = logger
        self.protPid = os.getpid()
        self.protocol.setPid(self.protPid)
        self.protocol._store(self.protocol._pid)
        self.prerequisites = list(map(int, self.protocol.getPrerequisites()))
        # Keep track of the last time the protocol was checked and
        # its modification date to avoid unnecessary db opening
        self.updatedProtocols = dict()
        self.initial_sleep = self._args.initial_sleep

    def getSleepTime(self):
        return self._args.sleepTime

    def getInitialSleepTime(self):
        return self.initial_sleep

    def _loadProtocol(self) -> Protocol:
        return getProtocolFromDb(self._args.projPath,
                                 self._args.dbPath,
                                 self._args.protId, chdir=True)

    def _log(self, msg):
        self.log.info(msg)

    def _updateProtocol(self, protocol):

        protId = protocol.getObjId()

        if protId in self.updatedProtocols:
            return self.updatedProtocols[protId]

        protDb = protocol.getDbPath()

        if os.path.exists(protDb):
            updateResult = self.project._updateProtocol(protocol)
            if updateResult == PROTOCOL_UPDATED:
                self._log("Updated protocol: %s (%s)" % (protId, protocol))
            else:
                self._log("The protocol %s (%s) is up to date" % (protId, protocol))
            self.updatedProtocols[protId] = protocol

        return protocol

    def _getProtocolFromPointer(self, pointer):
        """
        The function return a protocol from an attribute

           A) When the pointer points to a protocol

           B) When the pointer points to another object (INDIRECTLY).
              - The pointer has an _extended value (new parameters
                configuration in the protocol)

           C) When the pointer points to another object (DIRECTLY).
              - The pointer has not an _extended value (old parameters
                configuration in the protocol)
        """
        output = pointer.get()
        if isinstance(output, Protocol):  # case A
            protocol = output
        else:
            if pointer.hasExtended():  # case B
                protocol = pointer.getObjValue()
            else:  # case C
                protocol = self.project.getNode(str(output.getObjParentId())).run
        return protocol

    def _getSecondsToWait(self, inProt):
        """
        Assigns a timeout penalty or reward depending on the status of the
        input protocol (inProt)
        If inProt status is stopped we assign a reward i.o.c a penalty
        """
        protStatus = inProt.getStatus()
        inStreaming = inProt.worksInStreaming()
        meStreaming = self.protocol.worksInStreaming()

        penaltyRewardValues = {
            STATUS_LAUNCHED: 5 if inStreaming else 10,
            STATUS_RUNNING: -2 if inStreaming else 3,
            STATUS_SCHEDULED: int(self.getSleepTime()/2),
            STATUS_SAVED: self.getSleepTime(),
        }

        secondToWait = penaltyRewardValues.get(protStatus, -3)

        if not meStreaming:
            secondToWait += 3 * self.getSleepTime()

        return secondToWait

    def _checkPrerequisites(self, prerequisites, project):
        # Check if we need to wait for required protocols
        wait = False
        # For each protocol that is a prerequisite that has not finished,
        # we'll penalize with 3 more seconds of waiting
        penalize = 0
        self._log("Checking prerequisites... %s" % prerequisites)
        for protId in prerequisites:
            # Check if prerequisites exist. In the case of metaprotocols, it
            # may be necessary to load them from the project database.
            node = project.getRunsGraph().getNode(str(protId))

            if node is None:
                self._log("Updating runs graph. Missing protocol ... %s" % protId)
                project.getRunsGraph(refresh=True)

            node = project.getRunsGraph().getNode(str(protId))
            # Check if the protocol is within our workflow
            if node is None:
                self._log("Protocol can't wait for %s. Missing prerequisite " % protId)
                break

            prot = project.getRunsGraph().getNode(str(protId)).run
            if prot is not None:
                prot = self._updateProtocol(prot)
                penalize += self._getSecondsToWait(prot)
                if prot.getStatus() not in stopStatuses:
                    wait = True
                    self._log("   ...waiting for %s" % prot)
        return wait, penalize

    def _checkMissingInput(self):
        """
         Check if there are missing inputs
         In case of having them, check if protocol is in streaming
        """
        inputMissing = False
        # For each input that is not ready we'll penalize with 3 more
        # seconds of waiting
        penalize = 0

        self._log("Checking input data...")
        # Updating input protocols
        for key, attr in self.protocol.iterInputAttributes():
            self.log.debug("Turn for %s" % key)
            inputProt = self._getProtocolFromPointer(attr)
            inputProt = self._updateProtocol(inputProt)
            penalize += self._getSecondsToWait(inputProt)

        validation = self.protocol.validate()
        if len(validation) > 0:
            inputMissing = True
            self._log("%s doesn't validate:\n\t- %s" % (self.protocol.getObjLabel(),
                                                        '\n\t- '.join(
                                                            validation)))
        elif not self.protocol.worksInStreaming():
            for key, attr in self.protocol.iterInputAttributes():
                inSet = attr.get()
                if isinstance(inSet, Set) and inSet.isStreamOpen():
                    inputMissing = True
                    self._log("Waiting for closing %s... (does not work in "
                              "streaming)" % inSet)
                    break
                elif isinstance(inSet, Protocol) and not inSet.isFinished():  # Then is a pointer to a protocol
                    inputMissing = True
                    self._log("Waiting for protocol %s to finish... (does not work in "
                              "streaming)" % inSet)
                    break

                else:
                    self._log("%s is not blocking this protocol. All ready." % inSet)

        if not inputMissing:
            inputProtocolIds = self.protocol.getProtocolsToUpdate()
            for protId in inputProtocolIds:
                protocol = self.project.getProtocol(protId)
                self.log.debug("Turn from inputProtocolDict for %s" % protocol)
                self._updateProtocol(protocol)

        return inputMissing, penalize

    def schedule(self):
        self._log("Scheduling protocol %s, PID: %s,prerequisites: %s, works in streaming: %s." %
                  (self.protocol.getObjId(), self.protPid, self.prerequisites,
                   self.protocol.worksInStreaming()))

        initialSleepTime = runScheduler.getInitialSleepTime()
        self._log("Waiting %s seconds before start checking inputs " % initialSleepTime)
        time.sleep(initialSleepTime)

        while True:
            # Clear the list of protocols updated in the previous loop
            self.updatedProtocols.clear()
            sleepTime = self.getSleepTime()
            # FIXME: This does not cover all the cases:
            # When the user registers new coordinates after clicking the
            # "Analyze result" button, this action is registered in the project.sqlite
            # and not in it's own run.db and never gets updated. It is not critical and
            # will only affect a combination of json import with extended that will
            # appear after clicking on the "Analyze result" button.

            # Check if there are missing inputs
            missing, penalize = self._checkMissingInput()
            sleepTime += penalize

            # Check the prerequisites
            wait, penalize = self._checkPrerequisites(self.prerequisites,
                                                      self.project)
            sleepTime += penalize

            if not missing and not wait:
                break

            sleepTime = max(min(sleepTime, MAX_SLEEP_TIME), self.getSleepTime())

            self._log("Still not ready, sleeping %s seconds...\n" % sleepTime)
            time.sleep(sleepTime)

        self._log("Launching the protocol >>>>")
        self.project.launchProtocol(self.protocol, scheduled=True, force=True)

def parseArgs():
    parser = argparse.ArgumentParser()
    _addArg = parser.add_argument  # short notation

    _addArg("projPath", metavar='PROJECT_NAME',
            help="Project database path.")

    _addArg("dbPath", metavar='DATABASE_PATH',
            help="Protocol database path.")

    _addArg("protId", type=int, metavar='PROTOCOL_ID',
            help="Protocol ID.")

    _addArg("logPath", metavar='LOG_PATH', help='Path to the log file.')

    _addArg("--initial_sleep", type=int, default=0,
            dest='initial_sleep', metavar='SECONDS',
            help="Initial sleeping time (in seconds)")

    _addArg("--sleep_time", type=int, default=15,
            dest='sleepTime', metavar='SECONDS',
            help="Sleeping time (in seconds) between updates.")

    _addArg("--wait_for", nargs='*', type=int, default=[],
            dest='waitProtIds', metavar='PROTOCOL_ID',
            help="List of protocol ids that should be not running "
                 "(i.e, finished, aborted or failed) before this "
                 "run will be executed.")

    return parser.parse_args()


if __name__ == '__main__':

    # Create a child process
    # using os.fork() method
    pid = os.fork()

    # pid greater than 0 represents
    # the parent process
    if pid > 0:
        sys.exit(0)
    else:
        try:
            # Parse arguments
            args = parseArgs()

            # Configure logging
            LoggingConfigurator.setUpProtocolSchedulingLog(args.logPath)
            logger = logging.getLogger(__name__)

            runScheduler = RunScheduler(args, logger)
            runScheduler.schedule()

        except Exception as ex:
            print("Scheduling failed with these parameters: ", sys.argv)
            raise ex from None


# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Jose Gutierrez (jose.gutierrez@cnb.csic.es)
# *              Adrian Quintana (aquintana@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module defines the text used in the application.
"""


class Message:
    # Example Usage: 
    # MyMessage = Message()
    # print MyMessage.label

    # Header List
    VIEW_PROJECTS = 'Projects'
    VIEW_PROTOCOLS = 'Protocols'
    VIEW_DATA = 'Data'
    VIEW_UPLOAD = 'Upload'
    
    # Projects Template
    LABEL_PROJECTS = 'Projects'
    LABEL_CREATE_PROJECT = 'Create Project'
    LABEL_IMPORT_PROJECT = 'Import project'
    TITLE_CREATE_PROJECT = 'Enter the project name'
    TITLE_CREATE_PROJECT_NAME = 'Project Name: '
    TITLE_EDIT_OBJECT = 'Edit Object properties'
    MESSAGE_DELETE_PROJECT = 'This will *delete* the project and all its *data*. Are you sure?'
    LABEL_DELETE_PROJECT = '[Delete]'
    TITLE_DELETE_PROJECT = 'Confirm project deletion'
    LABEL_RENAME_PROJECT = '[Rename]'
    TITLE_RENAME_PROJECT = 'Confirm project renaming'
    LABEL_CREATED = 'Created: '
    LABEL_MODIFIED = 'Modified: '
    
    # Project Content Template
    LABEL_PROJECT = 'Project '

    # -- Protocol Treeview --
    LABEL_WORKFLOW = 'Workflow View: '
    LABEL_PROTTREE_NONE = 'None'

    # -- Toolbar --
    LABEL_NEW = 'New'
    LABEL_NEW_ACTION = 'New     '
    LABEL_EDIT = 'Edit'
    LABEL_RENAME = 'Rename '
    LABEL_EDIT_ACTION = 'Edit     '
    LABEL_COPY = 'Copy'
    LABEL_COPY_ACTION = 'Copy   '
    LABEL_DELETE = 'Delete'
    LABEL_DELETE_ACTION = 'Delete    '
    LABEL_STEPS = 'Steps'
    LABEL_BROWSE = 'Browse'
    LABEL_BROWSE_ACTION = 'Browse '
    LABEL_DB = 'Db'
    LABEL_STOP = 'Stop'
    LABEL_STOP_ACTION = 'Stop execution'
    LABEL_ANALYZE = 'Analyze Results'
    LABEL_TREE = 'Tree'
    LABEL_SMALLTREE = 'Small Tree'
    LABEL_REFRESH = 'Refresh'
    LABEL_DEFAULT = 'Default'
    LABEL_CONTINUE = 'Continue'
    LABEL_CONTINUE_ACTION = 'Approve continue'
    LABEL_EXPORT = 'Export'
    LABEL_EXPORT_UPLOAD = 'Export & upload'
    LABEL_RESTART_WORKFLOW = 'Restart workflow'
    LABEL_CONTINUE_WORKFLOW = 'Continue workflow'
    LABEL_STOP_WORKFLOW = 'Stop from here'
    LABEL_RESET_WORKFLOW = 'Reset from here'

    # -- Tabs --
    LABEL_DATA = 'Data'
    LABEL_SUMMARY = 'Summary'
    LABEL_INPUT = 'Input'
    LABEL_OUTPUT = 'Output'
    LABEL_COMMENT = 'Comments'
    
    LABEL_OBJSUMMARY = 'Object Summary'
    LABEL_OBJINFO = 'Info'
    LABEL_OBJCREATED = 'Created'
    LABEL_OBJLABEL = 'Label'
    
    LABEL_METHODS = 'Methods'
    LABEL_BIB_BTN = 'Export references'
    LABEL_LOGS = 'Output Logs'
    LABEL_LOGS_OUTPUT = 'Output Log'
    LABEL_LOGS_ERROR = 'Error Log'
    LABEL_LOGS_SCIPION = 'Scipion Log'
    
    NO_INFO_SUMMARY = 'No summary information.'
    NO_INFO_METHODS = 'No methods information.'
    NO_INFO_LOGS = 'No logs information.'
    NO_SAVE_SETTINGS = 'Error trying to save settings.'
    
    # ------- Protocol Form messages ----------
    LABEL_CITE = 'Cite'
    LABEL_HELP = 'Help'
    TEXT_HELP = 'The file selected will be uploaded to the project folder. If the file was uploaded before, it will be replaced.'
    LABEL_RUNNAME = 'Run name'
    LABEL_EXECUTION = 'Run mode'
    LABEL_RUNMODE = 'Mode'
    LABEL_PARALLEL = 'Parallel'
    LABEL_HOST = 'Host'
    LABEL_THREADS = 'Threads'
    LABEL_MPI = 'MPI'
    LABEL_QUEUE = 'Use a queue engine?'
    LABEL_QUEUE_FOR_JOBS = 'Use queue for Jobs?'

    LABEL_WAIT_FOR = 'Wait for'
    
    LABEL_EXPERT = 'Expert Level'
    LABEL_EXPERT_NORMAL = 'Normal'
    LABEL_EXPERT_ADVANCE = 'Advanced'
    LABEL_EXPERT_EXPERT = 'Expert'
    
    HELP_RUNMODE = """  
Normally, each protocol is composed of several atomic steps.
Each step could be computationally intensive, that's why
the *Continue* execution mode will try to continue from the
last completed step. On the other hand, the *Restart* mode
will clean the whole run directory and start from scratch.    
    """    

    HELP_MPI_THREADS = """
Define the number of processors to be used in the execution.
*MPI*: This is a number of independent processes
       that communicate through message passing
       over the network (or the same computer).
*Threads*: This refers to different execution threads 
       in the same process that can share memory. They run in
       the same computer.     
    """

    HELP_USEQUEUE = """
    Click Yes if you want to send this execution to a queue engine like Slurm, Torque, ...
    The queue commands to launch and stop jobs should be configured at
    _%s_ file.
    
    See %s for more information.
        """
    HELP_USEQUEUEPERJOB = """
    Click *Yes* if you want to submit the multiple jobs per protocol to a Queue system.
    The queue commands for launch and stop jobs should be configured
    for the current host in the _hosts.conf_ file.
    """

    HELP_WAIT_FOR = """
    Specify a comma separated list of protocol IDs if you want
    to *schedule* this protocol and wait for those protocols to finish before
    starting this one.
    
    This function will allow you to "schedule" many
    runs that will be executed after each other.
     
    See %s for more information.
    """
    
    TITLE_NAME_RUN = ' Protocol Run: '
    TITLE_RUN = 'Run'
    TITLE_LABEL = 'Label'
    LABEL_OPT_COMMENT = 'Describe your run here...'
    TITLE_COMMENT = 'Comment'
    LABEL_RUN_MODE_RESUME = 'resume'
    LABEL_RUN_MODE_RESTART = 'restart'
    TITLE_EXEC = 'Execution'
    TITLE_BROWSE_DATA = 'Protocol data'
    LABEL_QUEUE_YES = 'Yes'
    LABEL_QUEUE_NO = 'No'
    LABEL_PARAM_YES = 'Yes'
    LABEL_PARAM_NO = 'No'
    LABEL_BUTTON_CLOSE = 'Close'
    LABEL_BUTTON_SAVE = 'Save'
    LABEL_BUTTON_EXEC = 'Execute'
    LABEL_BUTTON_VIS = 'Visualize'
    LABEL_BUTTON_WIZ = 'Wizard'
    LABEL_BUTTON_HELP = 'Help'
    LABEL_BUTTON_RETURN = 'Save'
    # VARS
    VAR_EXEC_HOST = 'hostName'
    VAR_EXPERT = 'expertLevel'
    VAR_MPI = 'numberOfMpi'
    VAR_QUEUE = '_useQueue'
    VAR_RUN_NAME = 'runName'
    VAR_COMMENT = 'comment'
    VAR_RUN_MODE = 'runMode'
    VAR_THREADS = 'numberOfThreads'
    
    LABEL_PATTERN = 'Pattern'
    TEXT_PATTERN = """\
Pattern (that can include wildcards) of the files to import.
For example:
  *data/particles/***.spi*
  *~/Micrographs/mic/***.mrc*"""
    ERROR_PATTERN_EMPTY = 'The *pattern* cannot be empty.'
    ERROR_PATTERN_FILES = 'There are no files matching the *pattern*'
    LABEL_CHECKSTACK = 'Check stack files?'
    LABEL_COPYFILES = 'Copy files?'
    LABEL_VOLTAGE = 'Microscope voltage (kV)'
    TEXT_VOLTAGE = "Microscope voltage"
    LABEL_SPH_ABERRATION = 'Spherical aberration (mm)'
    TEXT_SPH_ABERRATION = """\
Optical effect due to the increased refraction of light rays when they
strike the lens near its edge, in comparison with those that strike near
the center."""
    LABEL_AMPLITUDE = 'Amplitude Contrast'
    TEXT_AMPLITUDE = """\
Produced by the loss of amplitude (i.e. electrons) from the beam.

For a weak phase and weak amplitude object, the amplitude contrast ratio Qo
is automatically computed. It should be a positive number, typically between
0.05 and 0.3."""
    LABEL_PATTERNU = 'Pattern untilted'
    LABEL_PATTERNT = 'Pattern tilted'

    LABEL_SAMP_MODE = 'Sampling rate mode'
    TEXT_SAMP_MODE = """\
You can specify the sampling rate (pixel size) directly from the image
(A/pixel, Ts) or by specifying the magnification rate (M) and the scanner
pixel size (microns/pixel, Tm).

They are related by  Ts = Tm / M"""
    LABEL_SAMP_MODE_1 = 'From image'
    LABEL_SAMP_MODE_2 = 'From scanner'
    LABEL_SAMP_RATE = 'Pixel size (sampling rate) Å/px'
    TEXT_SAMP_RATE = "Pixel size"
    LABEL_MAGNI_RATE = 'Magnification rate'
    TEXT_MAGNI_RATE = """\
Electron optical magnification (M). It can be used to compute the Image Pixel
Size ("Sampling Rate") (Ts) using the Scanner Pixel Size (Tm), Ts = Tm / M."""
    LABEL_SCANNED = 'Scanned pixel size (microns/px)'

    ERROR_IMPORT_VOL = 'importVolumes: There are no filePaths matching the pattern'
    
    LABEL_CTF_ESTI = 'CTF Estimation'
    LABEL_INPUT_MIC = 'Input Micrographs'
    LABEL_INPUT_PART = 'Input Particles'
    LABEL_INPUT_VOLS = 'Input Volumes'
    LABEL_INPUT_MOVS = 'Input Movies'
    LABEL_ALIG_PART = 'Write aligned particles?'
    TEXT_ALIG_PART = 'If set to *Yes*, the alignment will be applied to \n'+'input particles and a new aligned set will be created.'

    TEXT_NO_INPUT_MIC = 'No *Input Micrographs* selected.'
    TEXT_NO_CTF_READY = 'CTF of *Input Micrographs* not ready yet.'
    TEXT_NO_OUTPUT_CO = 'Output coordinates not ready yet.'
    ERROR_NO_EST_CTF = '_estimateCTF should be implemented'
    TEXT_NO_OUTPUT_FILES = 'No output file produced'
    
    TITLE_LAUNCHED = 'Success'
    LABEL_LAUNCHED = 'The protocol was launched successfully.'
    LABEL_FOUND_ERROR = 'Errors found'
    TITLE_SAVED_FORM = 'Success'
    LABEL_SAVED_FORM = 'The protocol was saved successfully.'
    TITLE_DELETE_FORM = 'Confirm DELETE'
    TITLE_RESTART_FORM = 'Confirm RESTART'
    LABEL_DELETE_FORM = """
You are going to *DELETE* the run(s): 
  - %s
*ALL DATA* related will be permanently removed.

Do you really want to continue?'
"""
    LABEL_RESTART_FORM = """
You are going to *RESTART* the run: 
  %s

Do you really want to restart it?'
"""
    LABEL_RESTART_FORM_MANY = """
Following data will be *REMOVED* : 
  %s

What do you want to restart?

  *Single* : Just this protocol, rest will be reset.
  *All*: All above listed protocols will be restarted.
"""

    TITLE_STOP_FORM = 'Confirm STOP'
    LABEL_STOP_FORM = 'Do you really want to *STOP* this run?'
    
    NO_VIEWER_FOUND = 'There is no viewer for protocol:'
    
    TITLE_SAVE_OUTPUT = 'Save protocol output'
    LABEL_SAVE_OUTPUT = 'Do you wish to save protocol output?'

    TITLE_RESTART_WORKFLOW_FORM = 'Confirm RESTART'
    TITLE_STOP_WORKFLOW_FORM = 'Confirm STOP'
    TITLE_RESET_WORKFLOW_FORM = 'Confirm RESET'
    TITLE_RESTART_WORKFLOW = 'Do you really want to *RESTART* this Workflow? All previous results will be deleted'
    TITLE_STOP_WORKFLOW = 'Do you really want to *STOP* this Workflow?'
    TITLE_RESET_WORKFLOW = 'Do you really want to *RESET* this Workflow?'
    TITLE_LAUNCHED_WORKFLOW_FAILED_FORM = 'Error while launching the workflow'
    TITLE_STOPPED_WORKFLOW_FAILED = 'Error while stopping the workflow'
    TITLE_RESETED_WORKFLOW_FAILED = 'Error while resetting the workflow'
    TITLE_LAUNCHED_WORKFLOW_FAILED = 'The workflow can not be relaunched from this protocol.\n'
    TITLE_ACTIVE_PROTOCOLS = 'The following protocols are active:'
    
    # SHOWJ_WEB
    SHOWJ_TITLE = 'Showj'
    
    LABEL_RESLICE = 'Reslice'
    RESLICE_Z_NEGATIVE = 'Z Negative (Front)'
    RESLICE_Y_NEGATIVE = 'Y Negative (Top)'
    RESLICE_Y_POSITIVE = 'Y Positive (Bottom)'
    RESLICE_X_NEGATIVE = 'X Negative (Left)'
    RESLICE_X_POSITIVE = 'X Positive (Right)'
    
    LABEL_COLS = 'Cols'
    LABEL_ROWS = 'Rows'
    
    LABEL_MIRRORY = 'Invert Y Axis'
    LABEL_APPLY_TRANSFORM = 'Apply Transform Matrix'
    LABEL_ONLY_SHIFTS = 'Only Shifts'
    LABEL_WRAP = 'Wrap'
    
    LABEL_BLOCK_SELECTION = 'Select Block'
    LABEL_LABEL_SELECTION = 'Select Label'
    LABEL_VOLUME_SELECTION = 'Select Volume'
    
    LABEL_ENABLE = 'Enable'
    LABEL_DISABLE = 'Disable'
    LABEL_SELECT_ALL = 'Select all'
    LABEL_SELECT_FROM = 'Select from here'
    LABEL_SELECT_TO = 'Select to here'
    
    LABEL_DISPLAY_TABLE_CONFIG = 'Display Table Configuration'
    
    LABEL_LABEL = 'Label'
    LABEL_VISIBLE = 'Visible'
    LABEL_RENDER = 'Render'
    
    LABEL_BUTTON_OK = 'Ok'
    LABEL_BUTTON_CANCEL = 'Cancel'
    
    LABEL_THRESHOLD = 'Threshold:'
    
    ERROR_DIMENSIONS = 'Incorrect table width or height: '
    ERROR_WEBGL = 'Your web browser does not support or is not configured for WebGL. See [[http://get.webgl.org/][WebGL Support]] for more information.'
    
    TOOLTIP_SEARCH = 'Search for a given world in the text. '
    TOOLTIP_SEARCH_NEXT = 'Move to the next highlighted item. Also, press <Down> or <F3>'
    TOOLTIP_SEARCH_PREVIOUS = 'Move to the previous highlighted item. Also, press <Up> or <Shift-F3>'
    TOOLTIP_REFRESH = 'Reload the content of the files in the viewer. '
    TOOLTIP_EXTERNAL = 'Open the viewer in an external window. '

    TITLE_PICK_GAUSS = 'Automatic gaussian picking'
    LABEL_PICK_GAUSS = 'Do you wish to perform an automatic gaussian picking for the remaining micrographs?'

    TITLE_INSPECTOR = 'Objects inspector'
    LABEL_INSPECTOR = 'Objects inspector will inspect the whole project. ' \
                      'Thus, it can take a while depending on the size of the project.\n' \
                      'Do you want to continue?'
    EXECUTE_PLUGINS_MANAGER_OPERATION = 'Execute all selected operations'
    CANCEL_SELECTED_OPERATION = 'Cancel a selected operation'


# PLUGIN/BINARY STATES
class PluginStates:
    PLUGIN = 'plugin'
    BINARY = 'binary'
    UNCHECKED = 'unchecked'
    CHECKED = 'checked'
    INSTALL = 'install'
    UNINSTALL = 'uninstall'
    TO_INSTALL = 'to_install'
    INSTALLED = 'installed'
    PRECESSING = 'processing'
    FAILURE = 'failure'
    AVAILABLE_RELEASE = 'available_release'
    TO_UPDATE = 'to_update'
    SUCCESS = 'success'
    ERRORS = 'errors'
    WAITING = 'waiting'


class PluginInformation:
    PLUGIN_URL = 'pluginURL'
    PLUGIN_NAME = 'pluginName'
    PLUGIN_VERSION = 'pluginVersion'
    PLUGIN_RELEASE_DATE = 'pluginUploadedDate'
    PLUGIN_DESCRIPTION = 'pluginDescription'
    PLUGIN_AUTHORS = 'pluginAuthor'


# To get font awesome icons into png use: http://fa2png.io/
class Icon:
    # Protocols status
    PROT_DISABLED = 'prot_disabled.png'
    BETA = 'beta.png'
    NEW = 'new.png'
    PRODUCTION = 'production.png'
    UPDATED = 'updated.png'
    GROUP = 'class_obj.png'
    DEBUG = 'debug.png'
    DOWNLOAD = 'fa-download.png'
    FILE_BW = 'fa-file-o.png'
    FIND = 'binoculares.png'
    SELECT_ALL = 'workflow.png'
    ERROR = 'fa-times-circle_alert.png'
    INFO = 'fa-info-circle_alert.png'
    ALERT = 'fa-exclamation-triangle_alert.png'
    JAVA_FILE = 'file_java.png'
    PYTHON_FILE = 'file_python.png'
    # Project window icons
    RUNS_TREE = 'fa-sitemap.png'
    ACTION_NEW = 'fa-plus-circle.png'
    ACTION_EDIT = 'fa-pencil.png'
    ACTION_SELECT_FROM = 'fa-arrow-down.png'
    ACTION_SELECT_TO = 'fa-arrow-up.png'
    ACTION_COPY = 'fa-files-o.png'
    ACTION_DELETE = 'fa-trash-o.png'
    ACTION_REFRESH = 'fa-refresh.png'
    ACTION_RENAME = 'rename.png'
    # TODO: change action_steps icon - fa-codefork?
    ACTION_STEPS = 'fa-list-ul.png'
    ACTION_BROWSE = 'fa-folder-open.png'
    ACTION_DB = 'fa-database.png'
    ACTION_TREE = None
    ACTION_STOP = 'fa-stop.png'
    ACTION_CONTINUE = 'fa-play-circle-o.png'
    ACTION_STOP_WORKFLOW = 'fa-stop-workflow.png'
    ACTION_RESULTS = 'fa-eye.png'
    ACTION_CLOSE = 'fa-times.png'
    ACTION_SAVE = 'fa-save.png'
    ACTION_VISUALIZE = 'fa-eye.png'
    ACTION_WIZ = 'fa-magic.png'
    ACTION_HELP = 'fa-question-circle.png'
    ACTION_REFERENCES = 'fa-external-link.png'
    ACTION_EXPORT = 'fa-external-link.png'
    ACTION_EXPORT_UPLOAD = 'fa-upload.png'
    ACTION_SEARCH = 'fa-search.png'
    ACTION_EXECUTE = 'fa-cogs.png'
    ACTION_IN = 'fa-sign-in.png'
    ACTION_OUT = 'fa-sign-out.png'
    ACTION_FIND_NEXT = 'fa-next.png'
    ACTION_FIND_PREVIOUS = 'fa-previous.png'
    ACTION_COLLAPSE = 'fa-minus-square.png'
    ACTION_EXPAND ='fa-plus-square.png'
    # Host template
    BUTTON_SELECT = 'fa-check.png'
    BUTTON_CLOSE = 'fa-times.png'
    BUTTON_CANCEL = 'fa-ban.png'
    BUTTON_SAVE = ACTION_SAVE

    ARROW_UP = 'fa-arrow-up.png'
    TAGS = 'fa-tags.png'
    HOME = 'fa-home.png'
    LIGHTBULB = 'fa-lightbulb-o.png'
    PLUS_CIRCLE = 'fa-plus-circle.png'
    ROCKET = 'fa-rocket.png'
    NO_IMAGE_128 = 'no-image128.png'
    FOLDER = 'file_folder.png'
    FOLDER_LINK = 'file_folder_link.png'
    FILE = 'file_generic.png'
    FILE_LINK = 'file_generic_link.png'
    FOLDER_OPEN = 'fa-folder-open.png'
    DB = 'file_sqlite.png'
    TXT_FILE = 'file_text.png'

    SCIPION_ICON = 'scipion_icon.png'
    SCIPION_ICON_PROJ = SCIPION_ICON  # 'scipion_icon_proj.png'
    SCIPION_ICON_PROJS = SCIPION_ICON  # 'scipion_icon_projs.png'
    SCIPION_ICON_PROT = SCIPION_ICON  # 'scipion_icon_prot.png'

    # PLUGIN MANAGER ICONS
    CHECKED = 'fa-checked.png'
    UNCHECKED = 'fa-unchecked.png'
    INSTALL = 'fa-install.png'
    UNINSTALL = 'fa-uninstall.png'
    TO_INSTALL = 'fa-to_install.png'
    INSTALLED = 'fa-installed.png'
    PROCESSING = 'fa-processing.png'
    FAILURE = 'fa-failure.png'
    DELETE_OPERATION = 'fa-delete-operation.png'
    TO_UPDATE = 'fa-update.png'
    WAITING = 'wait.gif'
    ACTION_UNDO = 'fa-undo.png'

    PLUGIN_AUTHORS = 'users.png'
    PLUGIN_DESCRIPTION = 'file_stack.png'
    PLUGIN_RELEASE_DATE = 'fa-upload.png'
    PLUGIN_VERSION = 'file_vol.png'
    PLUGIN_PACKAGE = 'file_folder.png'


class Color:
    RED_COLOR = 'Firebrick'  # Red color for background label  = #B22222
    LIGHT_RED_COLOR = '#F3CBCB'  # Very light red for row selection
    LIGHT_BLUE_COLOR = '#EAEBFF'  # Very light blue for even rows
    LIGHT_GREY_COLOR = '#EAEBEC'  # Light grey for background color in form, protocol, table header and west container
    LIGHT_GREY_COLOR_2 = '#F2F2F2'  # Very light grey for odd rows, input background, etc
    DARK_GREY_COLOR = '#6E6E6E'  # Very dark grey for project title, tubes, etc
    
    STATUS_SAVED = '#D9F1FA', 
    STATUS_LAUNCHED = '#D9F1FA', 
    STATUS_RUNNING = '#FCCE62', 
    STATUS_FINISHED = '#D2F5CB', 
    STATUS_FAILED = '#F5CCCB', 
    STATUS_INTERACTIVE = '#F3F5CB',
    STATUS_ABORTED = '#F5CCCB',


class colorText:
    """printing in colors, bold, etc,
       example: print colorText.BOLD + 'Hello World !' + color.END
    """
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


class KEYSYM:
    """ Keysym values for evaluating key pressed within events
    as reported at http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/key-names.html
    """
    DELETE = 'Delete'
    RETURN = 'Return'



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This modules contains classes required for the workflow
execution and tracking like: Step and Protocol
"""

# ------------------ Constants values -----------------------------------------

# Possible status of a protocol run, used mainly to monitor progress

STATUS_SAVED = "saved"  # Parameters saved for later use
STATUS_LAUNCHED = "launched"  # launched to queue system, only useful for protocols
STATUS_NEW = "new"
STATUS_RUNNING = "running"    # currently executing
STATUS_FAILED = "failed"      # it run and failed
STATUS_FINISHED = "finished"  # successfully finished
STATUS_ABORTED = "aborted"
STATUS_INTERACTIVE = "interactive"    # waiting for user interaction
STATUS_WAITING = "waiting"            # following status is used for streaming only
# Allow to schedule jobs not ready for running yet
STATUS_SCHEDULED = "scheduled"

ACTIVE_STATUS = [STATUS_LAUNCHED, STATUS_RUNNING, STATUS_INTERACTIVE,
                 STATUS_SCHEDULED]

# Execution modes
MODE_RESUME = 0    # Try to starting at the first changed step, skipping unchanged ones
MODE_RESTART = 1   # Restart the protocol from the beginning, deleting all previous results
MODE_CONTINUE = 2  # Continue from specific step, not widely used //DEPRECATED.
# JMRT: We now use 'Continue' label instead of 'Resume' which is more intuitive for users.
MODE_CHOICES = ('Continue', 'Restart')  # , 'Continue')

# Initial sleeping time (in seconds) in order to launch a scheduled protocol.
# This value is multiply to the protocol level into the workflow.
INITIAL_SLEEP_TIME = 30

# Maximum time (in seconds) waiting for a scheduled protocol to check if it
# can be launched.
MAX_SLEEP_TIME = 120


# Steps execution mode
STEPS_SERIAL = 0      # Execute steps serially, some of the steps can be mpi programs
STEPS_PARALLEL = 1    # Execute steps in parallel, through threads or mpi

# Level of expertise for the input parameters, mainly used in the protocol form
LEVEL_NORMAL = 0
LEVEL_ADVANCED = 1
LEVEL_CHOICES = ('Normal', 'Advanced')

# Param names for GPU processing
USE_GPU = 'useGpu'
GPU_LIST = 'gpuList'

# Job management
UNKNOWN_JOBID = -1

# File sizes
SIZE_1KB = 1024
SIZE_1MB = SIZE_1KB * SIZE_1KB
SIZE_1GB = SIZE_1MB * SIZE_1KB
SIZE_1TB = SIZE_1GB * SIZE_1KB



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              Jose Gutierrez (jose.gutierrez@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module is mainly for the Viewer class, which 
serve as base for implementing visualization tools(Viewer sub-classes).
"""

DESKTOP_TKINTER = 'tkinter'
WEB_DJANGO = 'django'


class Wizard(object):
    """ This is a special case of GUI to help the user selecting important parameters.

    The _targets attribute is used to define to which parameters the Wizard can deal with.
    It will be a list of tuples such as::

        _targets = [(DefImportMicrographs, ['voltage', sphericalAberration']), (DefCTFMicrographs, ['lowRes', 'highRes'])]

    """

    _targets = []
    _environments = [DESKTOP_TKINTER] # This can be ignored

    def show(self, form, *params):
        """
        EMPTY METHOD. Needs to be implemented in your class. This will be called to show the wizard.

        :param form: the protocol form, given access to to all parameters. Some times the same wizard will modify several elements in the form.
        :param params: a list of params to modify. Sometimes the wizard can be generic and can be used for different parameters in the same form.
        """
        pass
    
    def getView(self):
        """
        EMPTY METHOD. Deprecated.This method should return the string value of the view in web
        that will respond to this wizard. This method only should be implemented
        in those wizards that have WEB_DJANGO environment defined. 
        """
        return None



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


from pyworkflow.protocol import (STATUS_SAVED, STATUS_LAUNCHED, STATUS_RUNNING,
                                 STATUS_FINISHED, STATUS_FAILED,
                                 STATUS_INTERACTIVE, STATUS_ABORTED,
                                 STATUS_SCHEDULED)

from pyworkflow.utils.properties import Message, Icon


STATUS_COLORS = {
               STATUS_SAVED: '#D9F1FA',
               STATUS_LAUNCHED: '#D9F1FA',
               STATUS_RUNNING: '#FCCE62',
               STATUS_FINISHED: '#D2F5CB',
               STATUS_FAILED: '#F5CCCB',
               STATUS_INTERACTIVE: '#F3F5CB',
               STATUS_ABORTED: '#F5CCCB',
               STATUS_SCHEDULED: '#F3F5CB'
               }

# For protocols with warnings
WARNING_COLOR = '#848484'


ACTION_EDIT = Message.LABEL_EDIT
ACTION_RENAME = Message.LABEL_RENAME
ACTION_SELECT_FROM = Message.LABEL_SELECT_FROM
ACTION_SELECT_TO = Message.LABEL_SELECT_TO
ACTION_COPY = Message.LABEL_COPY
ACTION_DELETE = Message.LABEL_DELETE
ACTION_REFRESH = Message.LABEL_REFRESH
ACTION_STEPS = Message.LABEL_STEPS
ACTION_BROWSE = Message.LABEL_BROWSE
ACTION_DB = Message.LABEL_DB
ACTION_TREE = Message.LABEL_TREE
ACTION_STOP = Message.LABEL_STOP
ACTION_DEFAULT = Message.LABEL_DEFAULT
ACTION_CONTINUE = Message.LABEL_CONTINUE
ACTION_RESULTS = Message.LABEL_ANALYZE
ACTION_EXPORT = Message.LABEL_EXPORT
ACTION_EXPORT_UPLOAD = Message.LABEL_EXPORT_UPLOAD
ACTION_SWITCH_VIEW = 'Switch_View'
ACTION_COLLAPSE = 'Collapse'
ACTION_EXPAND = 'Expand'
ACTION_LABELS = 'Labels'
ACTION_RESTART_WORKFLOW = Message.LABEL_RESTART_WORKFLOW
ACTION_CONTINUE_WORKFLOW = Message.LABEL_CONTINUE_WORKFLOW
ACTION_STOP_WORKFLOW = Message.LABEL_STOP_WORKFLOW
ACTION_RESET_WORKFLOW = Message.LABEL_RESET_WORKFLOW
ACTION_SEARCH = 'Search'

ActionIcons = {
    ACTION_EDIT: Icon.ACTION_EDIT,
    ACTION_SELECT_FROM: Icon.ACTION_SELECT_FROM,
    ACTION_SELECT_TO: Icon.ACTION_SELECT_TO,
    ACTION_COPY: Icon.ACTION_COPY,
    ACTION_DELETE: Icon.ACTION_DELETE,
    ACTION_REFRESH: Icon.ACTION_REFRESH,
    ACTION_RENAME: Icon.ACTION_RENAME,
    ACTION_STEPS: Icon.ACTION_STEPS,
    ACTION_BROWSE: Icon.ACTION_BROWSE,
    ACTION_DB: Icon.ACTION_DB,
    ACTION_TREE: None,  # should be set
    ACTION_STOP: Icon.ACTION_STOP,
    ACTION_CONTINUE: Icon.ACTION_CONTINUE,
    ACTION_RESULTS: Icon.ACTION_RESULTS,
    ACTION_EXPORT: Icon.ACTION_EXPORT,
    ACTION_EXPORT_UPLOAD: Icon.ACTION_EXPORT_UPLOAD,
    ACTION_COLLAPSE: Icon.ACTION_COLLAPSE,
    ACTION_EXPAND: Icon.ACTION_EXPAND,
    ACTION_LABELS: Icon.TAGS,
    ACTION_RESTART_WORKFLOW: Icon.ACTION_EXECUTE,
    ACTION_CONTINUE_WORKFLOW: Icon.ACTION_CONTINUE,
    ACTION_STOP_WORKFLOW: Icon.ACTION_STOP_WORKFLOW,
    ACTION_RESET_WORKFLOW: Icon.ACTION_REFRESH,
    ACTION_SEARCH: Icon.ACTION_SEARCH
}



# Variable related to the online services (workflow stats and repository) that Scipion reports to or communicate.
# These variables may change between Scipion versions but do not depend on the user or system

# Web that gathers protocol usage
SCIPION_STATS_SERVER = 'https://scipion.i2pc.es'
SCIPION_STATS_WORKFLOW_APP = SCIPION_STATS_SERVER + '/report_protocols/api/workflow/workflow/'

# Web that handles workflows
WORKFLOW_REPOSITORY_SERVER = 'https://workflows.scipion.i2pc.es/'
WORKFLOW_PROG_STEP1 = 'workflowProgStep1_add/'
WORKFLOW_PROG_STEP2 = 'workflowProgStep2_add/'



#!/usr/bin/env python
# To run only the tests in this file, use:
# python -m unittest test_canvas -v

import pyworkflow.gui.canvas
import tkinter
import math
from pyworkflow.tests import *


class TestCanvas(BaseTest):
    # IMPORTANT: Tk requires at least that DISPLAY is defined
    # hence in some environments (like buildbot) the test may fail,
    # check for the TclError exception

    _labels = [PULL_REQUEST]

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)

    def distance(self, c1, c2):
        return round(math.hypot(c2[0] - c1[0], c2[1] - c1[1]), 2)

    def allEqual(self, values):
        return not values or values.count(values[0]) == len(values)

    def allDifferent(self, values):
        return not values or len(values) == len(set(values))

    def test_connectorsCoords(self):
        try:
            root = tkinter.Tk()
            canvas = pyworkflow.gui.canvas.Canvas(root, width=800, height=600)
            tb1 = canvas.createTextbox("First", 100, 100, "blue")

            connectorsCoords = tb1.getConnectorsCoordinates()
            self.assertTrue(self.allDifferent(connectorsCoords))

            print(connectorsCoords)

            distances = {}
            for i in range(len(connectorsCoords) - 1):
                distances[i] = self.distance(connectorsCoords[i],
                                             connectorsCoords[i + 1])

                print(distances)
                self.assertTrue(self.allEqual(list(distances.values())))
                self.assertNotEqual(distances[0], 0)
        except tkinter.TclError as ex:
            print(ex)

    def test_closestConnectors(self):
        try:
            root = tkinter.Tk()
            canvas = pyworkflow.gui.canvas.Canvas(root, width=800, height=600)
            tb1 = canvas.createTextbox("Textbox1", 100, 100, "blue")
            tb2 = canvas.createTextbox("Textbox2", 300, 100, "blue")
            tb3 = canvas.createTextbox("Textbox3", 100, 300, "blue")

            tb1ConnectorsCoords = tb1.getConnectorsCoordinates()
            tb2ConnectorsCoords = tb2.getConnectorsCoordinates()
            tb3ConnectorsCoords = tb3.getConnectorsCoordinates()
            c1, c2 = pyworkflow.gui.canvas.findStrictClosestConnectors(tb1, tb2)
            c3, c4 = pyworkflow.gui.canvas.findStrictClosestConnectors(tb1, tb3)
            # tb1 and tb2 are aligned vertically. tb1 and tb3, horizontally.
            # So, their closest connectors must share one coordinate (y in case of c1&c2, x i n case of c3&c4)
            self.assertEqual(c1[1], c2[1])
            self.assertEqual(c3[0], c4[0])

        except tkinter.TclError as ex:
            print(ex)



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Pablo Conesa [1]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" This modules hosts gui code to visualize steps"""
import json
import tkinter as tk

import pyworkflow.protocol as pwprot
import  pyworkflow.gui as pwgui
import pyworkflow.utils as pwutils


from pyworkflow import TK
from pyworkflow.utils import Icon


class StepsTreeProvider(pwgui.tree.TreeProvider):
    """Create the tree elements for a Protocol run"""

    def __init__(self, stepsList):
        for i, s in enumerate(stepsList):
            if not s._index:
                s._index = i + 1

        self._stepsList = stepsList
        self.getColumns = lambda: [('Index', 50), ('Step', 200), ('Status', 150),
                                   ('Time', 150), ('Class', 100)]
        self._parentDict = {}

    def getObjects(self):
        return self._stepsList

    @staticmethod
    def getObjectInfo(obj):
        info = {'key': obj._index,
                'values': (str(obj), obj.getStatus(), pwutils.prettyDelta(obj.getElapsedTime()),
                           obj.getClassName())}
        return info

    @staticmethod
    def getObjectPreview(obj):

        args = json.loads(obj.argsStr.get())
        msg = "*Prerequisites*: %s \n" % str(obj._prerequisites)
        msg += "*Arguments*: " + '\n  '.join([str(a) for a in args])
        if hasattr(obj, 'resultFiles'):
            results = json.loads(obj.resultFiles.get())
            if len(results):
                msg += "\n*Result files:* " + '\n  '.join(results)

        return None, msg

class StepsWindow(pwgui.browser.BrowserWindow):
    def __init__(self, title, parentWindow, protocol, **args):
        self._protocol = protocol
        provider = StepsTreeProvider(protocol.loadSteps())
        pwgui.browser.BrowserWindow.__init__(self, title, parentWindow,
                                             weight=False, **args)
        # Create buttons toolbar
        self.root.columnconfigure(0, weight=1)
        self.root.rowconfigure(1, weight=1)

        self.fillToolBar()

        # Create and set browser
        browser = pwgui.browser.ObjectBrowser(self.root, provider,
                                              showPreviewTop=False)
        self.setBrowser(browser, row=1, column=0)

    def fillToolBar(self):
        # Tool bar
        toolbar = tk.Frame(self.root)
        toolbar.grid(row=0, column=0, sticky='nw', padx=5, pady=5)

        # Tree button
        btn = tk.Label(toolbar, text="Tree",
                       image=self.getImage(Icon.RUNS_TREE),
                       compound=tk.LEFT, cursor='hand2')
        btn.bind(TK.LEFT_CLICK, self._showTree)
        btn.grid(row=0, column=0, sticky='nw')

        # Reset status
        btn = tk.Label(toolbar, text="Reset",
                       image=self.getImage(Icon.ACTION_REFRESH),
                       compound=tk.LEFT, cursor='hand2')
        btn.bind('<Button-1>', self._resetStep)
        btn.grid(row=0, column=1, sticky='nw')

    def _resetStep(self, e=None):

        item = self.browser._lastSelected
        if item is not None:
            objId = item.getObjId()
            self._protocol._updateSteps(lambda step: step.setStatus(pwprot.STATUS_NEW), where="id='%s'" % objId)
            item.setStatus(pwprot.STATUS_NEW)
            self.browser.tree.update()
    # noinspection PyUnusedLocal
    def _showTree(self, e=None):
        g = self._protocol.getStepsGraph()
        w = pwgui.Window("Protocol steps", self, minsize=(800, 600))
        root = w.root
        canvas = pwgui.Canvas(root, width=600, height=500,
                              tooltipCallback=self._stepTooltip,)
        canvas.grid(row=0, column=0, sticky='nsew')
        canvas.drawGraph(g, pwgui.LevelTreeLayout())
        w.show()

    def _stepTooltip(self, tw, item):
        """ Create the contents of the tooltip to be displayed
        for the given step.
        Params:
            tw: a tk.TopLevel instance (ToolTipWindow)
            item: the selected step.
        """

        if not hasattr(item.node, 'step'):
            return

        step = item.node.step

        tm = str(step.funcName)

        if not hasattr(tw, 'tooltipText'):
            frame = tk.Frame(tw)
            frame.grid(row=0, column=0)
            tw.tooltipText = pwgui.dialog.createMessageBody(
                frame, tm, None, textPad=0, textBg=pwutils.Color.LIGHT_GREY_COLOR_2)
            tw.tooltipText.config(bd=1, relief=tk.RAISED)
        else:
            pwgui.dialog.fillMessageText(tw.tooltipText, tm)





# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import logging
import threading
from pyworkflow import Config
from pyworkflow.gui import TextFileViewer, getDefaultFont, LIST_TREEVIEW, ShortCut
from pyworkflow.gui.project.constants import ACTION_REFRESH, ACTION_EDIT, ACTION_COPY, ACTION_DELETE, ACTION_STEPS, \
    ACTION_BROWSE, ACTION_DB, ACTION_STOP, ACTION_CONTINUE, ACTION_RESULTS, ACTION_EXPORT, ACTION_EXPORT_UPLOAD, \
    ACTION_COLLAPSE, ACTION_EXPAND, ACTION_LABELS, ACTION_SEARCH, ActionIcons, ACTION_TREE, ACTION_SWITCH_VIEW, \
    ACTION_SELECT_TO, ACTION_RENAME, ACTION_RESTART_WORKFLOW, ACTION_CONTINUE_WORKFLOW, ACTION_DEFAULT, \
    ACTION_SELECT_FROM, ACTION_STOP_WORKFLOW, ACTION_RESET_WORKFLOW
from pyworkflow.protocol import SIZE_1MB, SIZE_1GB, SIZE_1TB

INIT_REFRESH_SECONDS = Config.SCIPION_GUI_REFRESH_INITIAL_WAIT

"""
View with the protocols inside the main project window.
"""

import os
import json
import re
import tempfile
from collections import OrderedDict
import tkinter as tk
import tkinter.ttk as ttk
import datetime as dt

from pyworkflow import Config, TK
import pyworkflow.utils as pwutils
import pyworkflow.protocol as pwprot
from pyworkflow.viewer import DESKTOP_TKINTER, ProtocolViewer
from pyworkflow.utils.properties import Color, KEYSYM, Icon, Message
from pyworkflow.webservices import WorkflowRepository

import pyworkflow.gui as pwgui
from pyworkflow.gui.form import FormWindow
from pyworkflow.gui.project.utils import getStatusColorFromNode, inspectObj
from pyworkflow.gui.project.searchprotocol import SearchProtocolWindow, ProtocolTreeProvider
from pyworkflow.gui.project.steps import StepsWindow
from pyworkflow.gui.project.viewprotocols_extra import RunIOTreeProvider, ProtocolTreeConfig
from pyworkflow.gui.project.searchrun import RunsTreeProvider, SearchRunWindow

logger = logging.getLogger(__name__)

DEFAULT_BOX_COLOR = '#f8f8f8'


RUNS_TREE = Icon.RUNS_TREE

VIEW_LIST = 0
VIEW_TREE = 1
VIEW_TREE_SMALL = 2


class ScipionLogWindow(pwgui.Window):
    """Class that create a windows where the system log is display """
    def __init__(self, parentWindow, **kwargs):
        pwgui.Window.__init__(self, title="Scipion log",
                              masterWindow=parentWindow,
                              minsize=(1000, 400))
        content = tk.Frame(self.root)
        content.grid(row=0, column=0, sticky='news')
        pwgui.configureWeigths(content)
        self.showScipionLog = threading.Thread(name="scipion_log",
                                              target=self._showScipionLog,
                                              args=(content,))
        self.showScipionLog.start()

    def _showScipionLog(self, content):
        """
        Create a content of the system log window
        """

        # Fill the Output Log
        terminal = tk.Frame(content)
        terminal.grid(row=0, column=0, sticky='news')
        pwgui.configureWeigths(terminal)

        self.textLog = TextFileViewer(terminal, font=getDefaultFont(),
                                      height=30, width=100)
        self.textLog.grid(row=0, column=0, sticky='news')

        fileLogPath = Config.SCIPION_LOG
        self.fileLog = open(fileLogPath, 'r')
        # Create a tab where the log will appear
        self.textLog.createWidgets([fileLogPath])
        self.textLog.refreshAll(goEnd=True)
        # Refreshing the log every 3 seconds
        self.threadRefresh = threading.Thread(name="refresh_log",
                                              target=self._refreshLogComponent,
                                              args=(3,))
        self.threadRefresh.start()

    def _refreshLogComponent(self, wait=3):
        """ Refresh the Plugin Manager log """
        import time
        while True:
            time.sleep(wait)
            # Taking the vertical scroll position. If this action fail, assume
            # that the log window was closed and finalized the refresh thread
            try:
                vsPos = self.textLog.taList[0].getVScroll()
                if vsPos[1] == 1.0:
                    self.textLog.refreshAll(goEnd=True)
                else:
                    self.textLog.refreshAll(goEnd=False)
            except Exception:
                break

# noinspection PyAttributeOutsideInit
class ProtocolsView(tk.Frame):
    """ What you see when the "Protocols" tab is selected.

    In the main project window there are three tabs: "Protocols | Data | Hosts".
    This extended tk.Frame is what will appear when Protocols is on.
    """

    RUNS_CANVAS_NAME = "runs_canvas"

    SIZE_COLORS = {SIZE_1MB: "green",
                    SIZE_1GB: "orange",
                    SIZE_1TB: "red"}

    _protocolViews = None

    def __init__(self, parent, windows, **args):
        tk.Frame.__init__(self, parent, **args)
        # Load global configuration
        self.windows = windows
        self.project = windows.project
        self.domain = self.project.getDomain()
        self.root = windows.root
        self.getImage = windows.getImage
        self.protCfg = self.getCurrentProtocolView()
        self.settings = windows.getSettings()
        self.runsView = self.settings.getRunsView()
        self._loadSelection()
        self._items = {}
        self._lastSelectedProtId = None
        self._lastStatus = None
        self.selectingArea = False
        self._lastRightClickPos = None  # Keep last right-clicked position

        self.style = ttk.Style()
        self.root.bind("<F5>", self.refreshRuns)
        self.root.bind("<Control-f>", self._findProtocol)
        self.root.bind("<Control-a>", self._selectAllProtocols)
        self.root.bind("<Control-t>", self._toggleColorScheme)
        self.root.bind("<Control-d>", self._toggleDebug)
        self.root.bind("<Control-l>", self._locateProtocol)
        self.root.bind("<F2>", self._F2Pressed)

        if Config.debugOn():
            self.root.bind("<Control-i>", self._inspectProtocols)

        # To bind key press to methods
        # Listen to any key: send event to keyPressed method
        self.root.bind("<Key>", self.keyPressed)
        self.keybinds = dict()

        # Register key binds
        self._bindKeyPress(KEYSYM.DELETE, self._onDelPressed)

        self.__autoRefresh = None
        self.__autoRefreshCounter = INIT_REFRESH_SECONDS  # start by 3 secs

        self.refreshSemaphore = True
        self.repeatRefresh = False

        c = self.createContent()
        pwgui.configureWeigths(self)
        c.grid(row=0, column=0, sticky='news')

    def _bindKeyPress(self, key, method):

        self.keybinds[key] = method

    def keyPressed(self, event):

        if event.keysym in self.keybinds:
            method = self.keybinds[event.keysym]

            method()

    def createContent(self):
        """ Create the Protocols View for the Project.
        It has two panes:
            Left: containing the Protocol classes tree
            Right: containing the Runs list
        """
        p = tk.PanedWindow(self, orient=tk.HORIZONTAL, bg='white')
        bgColor = Color.LIGHT_GREY_COLOR
        # Left pane, contains Protocols Pane
        leftFrame = tk.Frame(p, bg=bgColor)
        leftFrame.columnconfigure(0, weight=1)
        leftFrame.rowconfigure(1, weight=1)

        # Protocols Tree Pane
        protFrame = tk.Frame(leftFrame, width=300, height=500, bg=bgColor)
        protFrame.grid(row=1, column=0, sticky='news', padx=5, pady=5)
        protFrame.columnconfigure(0, weight=1)
        protFrame.rowconfigure(1, weight=1)
        self._createProtocolsPanel(protFrame, bgColor)
        self.updateProtocolsTree(self.protCfg)
        # Create the right Pane that will be composed by:
        # a Action Buttons TOOLBAR in the top
        # and another vertical Pane with:
        # Runs History (at Top)

        # Selected run info (at Bottom)
        rightFrame = tk.Frame(p, bg='white')
        rightFrame.columnconfigure(0, weight=1)
        rightFrame.rowconfigure(1, weight=1)
        # rightFrame.rowconfigure(0, minsize=label.winfo_reqheight())

        # Create the Action Buttons TOOLBAR
        toolbar = tk.Frame(rightFrame, bg='white')
        toolbar.grid(row=0, column=0, sticky='news')
        pwgui.configureWeigths(toolbar)
        # toolbar.columnconfigure(0, weight=1)
        toolbar.columnconfigure(1, weight=1)

        self.runsToolbar = tk.Frame(toolbar, bg='white')
        self.runsToolbar.grid(row=0, column=0, sticky='sw')
        # On the left of the toolbar will be other
        # actions that can be applied to all runs (refresh, graph view...)
        self.allToolbar = tk.Frame(toolbar, bg='white')
        self.allToolbar.grid(row=0, column=10, sticky='se')
        self.createActionToolbar()

        # Create the Run History tree
        v = ttk.PanedWindow(rightFrame, orient=tk.VERTICAL)
        # runsFrame = ttk.Labelframe(v, text=' History ', width=500, height=500)
        runsFrame = tk.Frame(v, bg='white')
        # runsFrame.grid(row=1, column=0, sticky='news', pady=5)
        self.runsTree = self.createRunsTree(runsFrame)
        pwgui.configureWeigths(runsFrame)

        self.createRunsGraph(runsFrame)

        if self.runsView == VIEW_LIST:
            treeWidget = self.runsTree
        else:
            treeWidget = self.runsGraphCanvas

        treeWidget.grid(row=0, column=0, sticky='news')

        # Create the Selected Run Info
        infoFrame = tk.Frame(v)
        infoFrame.columnconfigure(0, weight=1)
        infoFrame.rowconfigure(1, weight=1)
        # Create the info label
        self.infoLabel = tk.Label(infoFrame)
        self.infoLabel.grid(row=0, column=0, sticky='w', padx=3)
        # Create the Analyze results button
        self.btnAnalyze = pwgui.Button(infoFrame, text=Message.LABEL_ANALYZE,
                                       fg='white', bg=Color.RED_COLOR,
                                       image=self.getImage(Icon.ACTION_VISUALIZE),
                                       compound=tk.LEFT,
                                       activeforeground='white',
                                       activebackground='#A60C0C',
                                       command=self._analyzeResultsClicked)
        self.btnAnalyze.grid(row=0, column=0, sticky='ne', padx=15)
        # self.style.configure("W.TNotebook")#, background='white')
        tab = ttk.Notebook(infoFrame)  # , style='W.TNotebook')

        # Summary tab
        dframe = tk.Frame(tab, bg='white')
        pwgui.configureWeigths(dframe, row=0)
        pwgui.configureWeigths(dframe, row=2)
        # Just configure the provider, later below, in updateSelection, it will be
        # provided with the protocols.
        provider = RunIOTreeProvider(self, None,
                                     self.project.mapper, self.info)

        rowheight = pwgui.getDefaultFont().metrics()['linespace']
        self.style.configure("NoBorder.Treeview", background='white',
                             borderwidth=0, font=self.windows.font,
                             rowheight=rowheight)
        self.infoTree = pwgui.browser.BoundTree(dframe, provider, height=6,
                                                show='tree',
                                                style="NoBorder.Treeview")
        self.infoTree.grid(row=0, column=0, sticky='news')
        label = tk.Label(dframe, text='SUMMARY', bg='white',
                         font=self.windows.fontBold)
        label.grid(row=1, column=0, sticky='nw', padx=(15, 0))

        hView = {'sci-open': self._viewObject,
                 'sci-bib': self._bibExportClicked}

        self.summaryText = pwgui.text.TaggedText(dframe, width=40, height=5,
                                                 bg='white', bd=0,
                                                 font=self.windows.font,
                                                 handlers=hView)
        self.summaryText.grid(row=2, column=0, sticky='news', padx=(30, 0))

        # Method tab
        mframe = tk.Frame(tab)
        pwgui.configureWeigths(mframe)
        # Methods text box
        self.methodText = pwgui.text.TaggedText(mframe, width=40, height=15,
                                                bg='white', handlers=hView)
        self.methodText.grid(row=0, column=0, sticky='news')
        # Reference export button
        # btnExportBib = pwgui.Button(mframe, text=Message.LABEL_BIB_BTN,
        #                             fg='white', bg=Color.RED_COLOR,
        #                             image=self.getImage(Icon.ACTION_BROWSE),
        #                             compound=tk.LEFT,
        #                             activeforeground='white',
        #                             activebackground='#A60C0C',
        #                             command=self._bibExportClicked)
        # btnExportBib.grid(row=2, column=0, sticky='w', padx=0)

        # Logs
        ologframe = tk.Frame(tab)
        pwgui.configureWeigths(ologframe)
        self.outputViewer = pwgui.text.TextFileViewer(ologframe, allowOpen=True,
                                                      font=self.windows.font)
        self.outputViewer.grid(row=0, column=0, sticky='news')
        self.outputViewer.windows = self.windows

        self._updateSelection()

        # Move to the selected protocol
        if self._isSingleSelection():
            prot = self.getSelectedProtocol()
            node = self.runsGraph.getNode(str(prot.getObjId()))
            self._selectNode(node)


        # Add all tabs

        tab.add(dframe, text=Message.LABEL_SUMMARY)
        tab.add(mframe, text=Message.LABEL_METHODS)
        tab.add(ologframe, text=Message.LABEL_LOGS_OUTPUT)
        #         tab.add(elogframe, text=Message.LABEL_LOGS_ERROR)
        #         tab.add(slogframe, text=Message.LABEL_LOGS_SCIPION)
        tab.grid(row=1, column=0, sticky='news')

        v.add(runsFrame, weight=1)
        v.add(infoFrame, weight=20)
        v.grid(row=1, column=0, sticky='news')

        # Add sub-windows to PanedWindows
        p.add(leftFrame, padx=0, pady=0, sticky='news')
        p.add(rightFrame, padx=0, pady=0)
        p.paneconfig(leftFrame, minsize=5)
        leftFrame.config(width=235)
        p.paneconfig(rightFrame, minsize=10)

        return p

    def _viewObject(self, objId):
        """ Call appropriate viewer for objId. """
        proj = self.project
        obj = proj.getObject(int(objId))
        viewerClasses = self.domain.findViewers(obj.getClassName(), DESKTOP_TKINTER)
        if not viewerClasses:
            return  # TODO: protest nicely
        viewer = viewerClasses[0](project=proj, parent=self.windows)
        viewer.visualize(obj)

    def _loadSelection(self):
        """ Load selected items, remove if some do not exists. """
        self._selection = self.settings.runSelection
        for protId in list(self._selection):

            if not self.project.doesProtocolExists(protId):
                self._selection.remove(protId)

    def _isMultipleSelection(self):
        return len(self._selection) > 1

    def _isSingleSelection(self):
        return len(self._selection) == 1

    def _noSelection(self):
        return len(self._selection) == 0

    def info(self, message):
        self.infoLabel.config(text=message)
        self.infoLabel.update_idletasks()

    def cleanInfo(self):
        self.info("")

    def refreshRuns(self, e=None, initRefreshCounter=True, checkPids=False):
        """
        Refresh the protocol runs workflow. If the variable REFRESH_WITH_THREADS
        exits, then use a threads to refresh, i.o.c use normal behavior
        """
        useThreads = Config.refreshInThreads()
        if useThreads:
            import threading
            # Refresh the status of displayed runs.
            if self.refreshSemaphore:
                # print("Launching a thread to refresh the runs...")
                threadRefreshRuns = threading.Thread(name="Refreshing runs",
                                                     target=self.refreshDisplayedRuns,
                                                     args=(e, initRefreshCounter,
                                                           checkPids))
                threadRefreshRuns.start()
            else:
                self.repeatRefresh = True
        else:
            self.refreshDisplayedRuns(e, initRefreshCounter, checkPids)

    # noinspection PyUnusedLocal
    def refreshDisplayedRuns(self, e=None, initRefreshCounter=True, checkPids=False):
        """ Refresh the status of displayed runs.
         Params:
            e: Tk event input
            initRefreshCounter: if True the refresh counter will be set to 3 secs
             then only case when False is from _automaticRefreshRuns where the
             refresh time is doubled each time to avoid refreshing too often.
        """
        self.viewButtons[ACTION_REFRESH]['state'] = tk.DISABLED
        self.info('Refreshing...')
        self.refreshSemaphore = False
        if Config.debugOn():
            import psutil
            proc = psutil.Process(os.getpid())
            mem = psutil.virtual_memory()
            logger.debug("------------- refreshing ---------- ")
            files = proc.open_files()
            logger.debug("  open files: %s" % len(files))
            for f in files:
                logger.debug("    - %s, %s" % (f.path, f.fd))
            logger.debug("  memory percent: %s" % proc.memory_percent())

        if self.runsView == VIEW_LIST:
            self.updateRunsTree(True)
        else:
            self.updateRunsGraph(True, checkPids=checkPids)
            self._updateSelection()

        if initRefreshCounter:

            self.__autoRefreshCounter = INIT_REFRESH_SECONDS  # start by 3 secs
            if self.__autoRefresh:
                self.runsTree.after_cancel(self.__autoRefresh)
                self.__autoRefresh = self.runsTree.after(
                    self.__autoRefreshCounter * 1000,
                    self._automaticRefreshRuns)
        self.refreshSemaphore = True
        if self.repeatRefresh:
            self.repeatRefresh = False
            self.refreshRuns()
        self.cleanInfo()
        self.viewButtons[ACTION_REFRESH]['state'] = tk.NORMAL

    # noinspection PyUnusedLocal
    def _automaticRefreshRuns(self, e=None):
        """ Schedule automatic refresh increasing the time between refreshes. """
        if pwutils.envVarOn(Config.SCIPION_GUI_CANCEL_AUTO_REFRESH):
            return

        self.refreshRuns(initRefreshCounter=False, checkPids=True)
        secs = self.__autoRefreshCounter
        # double the number of seconds up to 30 min
        self.__autoRefreshCounter = min(2 * secs, 1800)
        self.__autoRefresh = self.runsTree.after(secs * 1000,
                                                 self._automaticRefreshRuns)

    # noinspection PyUnusedLocal
    def _findProtocol(self, e=None):
        """ Find a desired protocol by typing some keyword. """
        window = SearchProtocolWindow(self.windows)
        window.show()

    def _locateProtocol(self, e=None):

        window = SearchRunWindow(self.windows, self.runsGraph, onDoubleClick=self._onRunClick)
        window.show()
        # self._moveCanvas(0,1)

    def _onRunClick(self, e=None):
        """ Callback to be called when a click happens o a run in the SearchRunWindow.tree"""
        tree = e.widget
        protId = tree.getFirst()
        node = self.runsGraph.getNode(protId)
        self._selectNode(node)

    def _selectNode(self, node):

        x = node.x
        y = node.y
        self._moveCanvas(x,y)

        # Select the protocol
        self._selectItemProtocol(node.run)
        self.refreshDisplayedRuns()

    def _moveCanvas(self, X, Y):

        self.runsGraphCanvas.moveTo(X, Y)

    def _scipionLog(self, e=None):
        windows = ScipionLogWindow(self.windows)
        windows.show()

    def createActionToolbar(self):
        """ Prepare the buttons that will be available for protocol actions. """

        self.actionButtons = {}
        self.actionList = [ACTION_EDIT, ACTION_COPY, ACTION_DELETE,
                           ACTION_STEPS, ACTION_BROWSE, ACTION_DB,
                           ACTION_STOP, ACTION_CONTINUE, ACTION_RESULTS,
                           ACTION_EXPORT, ACTION_EXPORT_UPLOAD, ACTION_COLLAPSE,
                           ACTION_EXPAND, ACTION_LABELS, ACTION_SEARCH]

        def addButton(action, text, toolbar):
            btn = tk.Label(toolbar, text=text,
                           image=self.getImage(ActionIcons.get(action, None)),
                           compound=tk.LEFT, cursor='hand2', bg='white')
            btn.bind(TK.LEFT_CLICK, lambda e: self._runActionClicked(action))
            return btn

        for action in self.actionList:
            self.actionButtons[action] = addButton(action, action,
                                                   self.runsToolbar)

        ActionIcons[ACTION_TREE] = RUNS_TREE

        self.viewButtons = {}

        # Add combo for switch between views
        viewFrame = tk.Frame(self.allToolbar)
        viewFrame.grid(row=0, column=0)
        self._createViewCombo(viewFrame)

        # Add refresh Tree button
        btn = addButton(ACTION_TREE, "  ", self.allToolbar)
        pwgui.tooltip.ToolTip(btn, "Re-organize the node positions.", 1500)
        self.viewButtons[ACTION_TREE] = btn
        if self.runsView != VIEW_LIST:
            btn.grid(row=0, column=1)

        # Add refresh button
        btn = addButton(ACTION_REFRESH, ACTION_REFRESH, self.allToolbar)
        btn.grid(row=0, column=2)
        self.viewButtons[ACTION_REFRESH] = btn

    def _createViewCombo(self, parent):
        """ Create the select-view combobox. """
        label = tk.Label(parent, text='View:', bg='white')
        label.grid(row=0, column=0)
        viewChoices = ['List', 'Tree', 'Tree - small']
        self.switchCombo = pwgui.widgets.ComboBox(parent, width=10,
                                                  choices=viewChoices,
                                                  values=[VIEW_LIST, VIEW_TREE, VIEW_TREE_SMALL],
                                                  initial=viewChoices[self.runsView],
                                                  onChange=lambda e: self._runActionClicked(
                                                      ACTION_SWITCH_VIEW))
        self.switchCombo.grid(row=0, column=1)

    def _updateActionToolbar(self):
        """ Update which action buttons should be visible. """

        def displayAction(actionToDisplay, column, condition=True):

            """ Show/hide the action button if the condition is met. """

            # If action present (set color is not in the toolbar but in the
            # context menu)
            action = self.actionButtons.get(actionToDisplay, None)
            if action is not None:
                if condition:
                    action.grid(row=0, column=column, sticky='sw',
                                padx=(0, 5), ipadx=0)
                else:
                    action.grid_remove()

        for i, actionTuple in enumerate(self.provider.getActionsFromSelection()):
            action, cond = actionTuple
            displayAction(action, i, cond)

    def _createProtocolsTree(self, parent,
                             show='tree', columns=None):

        t = pwgui.tree.Tree(parent, show=show, style=LIST_TREEVIEW,
                            columns=columns)
        t.column('#0', minwidth=300)

        def configureTag(tag, img):
            # Protocol nodes
            t.tag_configure(tag, image=self.getImage(img))
            t.tag_bind(tag, TK.LEFT_DOUBLE_CLICK, self._protocolItemClick)
            t.tag_bind(tag, TK.RETURN, self._protocolItemClick)
            t.tag_bind(tag, TK.ENTER, self._protocolItemClick)

        # Protocol nodes
        configureTag(ProtocolTreeConfig.TAG_PROTOCOL, Icon.PRODUCTION)
        # New protocols
        configureTag(ProtocolTreeConfig.TAG_PROTOCOL_NEW, Icon.NEW)
        # Beta protocols
        configureTag(ProtocolTreeConfig.TAG_PROTOCOL_BETA, Icon.BETA)
        # Disable protocols (not installed) are allowed to be added.
        configureTag(ProtocolTreeConfig.TAG_PROTOCOL_DISABLED,
                     Icon.PROT_DISABLED)
        # Updates protocols
        configureTag(ProtocolTreeConfig.TAG_PROTOCOL_UPDATED,
                     Icon.UPDATED)
        t.tag_configure('protocol_base', image=self.getImage(Icon.GROUP))
        t.tag_configure('protocol_group', image=self.getImage(Icon.GROUP))
        t.tag_configure('section', font=self.windows.fontBold)
        return t

    def _createProtocolsPanel(self, parent, bgColor):
        """Create the protocols Tree displayed in left panel"""
        comboFrame = tk.Frame(parent, bg=bgColor)
        tk.Label(comboFrame, text='View', bg=bgColor).grid(row=0, column=0,
                                                           padx=(0, 5), pady=5)
        choices = self.getProtocolViews()
        initialChoice = self.settings.getProtocolView()
        combo = pwgui.widgets.ComboBox(comboFrame, choices=choices,
                                       initial=initialChoice)
        combo.setChangeCallback(self._onSelectProtocols)
        combo.grid(row=0, column=1)
        comboFrame.grid(row=0, column=0, padx=5, pady=5, sticky='nw')

        t = self._createProtocolsTree(parent)
        t.grid(row=1, column=0, sticky='news')
        # Program automatic refresh
        t.after(3000, self._automaticRefreshRuns)
        self.protTree = t

    def getProtocolViews(self):

        if self._protocolViews is None:
            self._loadProtocols()

        return list(self._protocolViews.keys())

    def getCurrentProtocolView(self):
        """ Select the view that is currently selected.
        Read from the settings the last selected view
        and get the information from the self._protocolViews dict.
        """
        currentView = self.project.getProtocolView()
        if currentView in self.getProtocolViews():
            viewKey = currentView
        else:
            viewKey = self.getProtocolViews()[0]
            self.project.settings.setProtocolView(viewKey)
            if currentView is not None:
                print("PROJECT: Warning, protocol view '%s' not found." % currentView)
                print("         Using '%s' instead." % viewKey)

        return self._protocolViews[viewKey]

    def _loadProtocols(self):
        """ Load protocol configuration from a .conf file. """
        # If the host file is not passed as argument...
        configProtocols = Config.SCIPION_PROTOCOLS

        localDir = Config.SCIPION_LOCAL_CONFIG
        protConf = os.path.join(localDir, configProtocols)
        self._protocolViews = ProtocolTreeConfig.load(self.project.getDomain(),
                                                      protConf)

    def _onSelectProtocols(self, combo):
        """ This function will be called when a protocol menu
        is selected. The index of the new menu is passed. 
        """
        protView = combo.getText()
        self.settings.setProtocolView(protView)
        self.protCfg = self.getCurrentProtocolView()
        self.updateProtocolsTree(self.protCfg)

    def populateTree(self, tree, treeItems, prefix, obj, subclassedDict, level=0):

        # If node does not have leaves (protocols) do not add it
        if not obj.visible:
            return

        text = obj.text
        if text:
            value = obj.value if obj.value is not None else text
            key = '%s.%s' % (prefix, value)
            img = obj.icon if obj.icon is not None else ''
            tag = obj.tag if obj.tag is not None else ''

            if len(img):
                img = self.getImage(img)
                # If image is none
                img = img if img is not None else ''

            protClassName = value.split('.')[-1]  # Take last part
            emProtocolsDict = self.domain.getProtocols()
            prot = emProtocolsDict.get(protClassName, None)

            if tag == 'protocol' and text == 'default':
                if prot is None:
                    print("Protocol className '%s' not found!!!. \n"
                          "Fix your config/protocols.conf configuration."
                          % protClassName)
                    return

                text = prot.getClassLabel()

            item = tree.insert(prefix, 'end', key, text=text, image=img, tags=tag)
            treeItems[item] = obj
            # Check if the attribute should be open or close
            openItem = getattr(obj, 'openItem', level < 2)
            if openItem:
                tree.item(item, open=openItem)

            # I think this mode is deprecated
            if obj.value is not None and tag == 'protocol_base':
                logger.warning('protocol_base tags are deprecated')
        else:
            key = prefix

        for sub in obj:
            self.populateTree(tree, treeItems, key, sub, subclassedDict,
                              level + 1)

    def updateProtocolsTree(self, protCfg):

        try:
            self.protCfg = protCfg
            self.protTree.clear()
            self.protTree.unbind(TK.TREEVIEW_OPEN)
            self.protTree.unbind(TK.TREEVIEW_CLOSE)
            self.protTreeItems = {}
            subclassedDict = {}  # Check which classes serve as base to not show them
            emProtocolsDict = self.domain.getProtocols()
            for _, v1 in emProtocolsDict.items():
                for k2, v2 in emProtocolsDict.items():
                    if v1 is not v2 and issubclass(v1, v2):
                        subclassedDict[k2] = True
            self.populateTree(self.protTree, self.protTreeItems, '', self.protCfg,
                              subclassedDict)
            self.protTree.bind(TK.TREEVIEW_OPEN,
                               lambda e: self._treeViewItemChange(True))
            self.protTree.bind(TK.TREEVIEW_CLOSE,
                               lambda e: self._treeViewItemChange(False))
        except Exception as e:
            # Tree can't be loaded report back, but continue
            print("Protocols tree couldn't be loaded: %s" % e)

    def _treeViewItemChange(self, openItem):
        item = self.protTree.focus()
        if item in self.protTreeItems:
            self.protTreeItems[item].openItem = openItem

    def createRunsTree(self, parent):
        self.provider = RunsTreeProvider(self.project, self._runActionClicked)

        # This line triggers the getRuns for the first time.
        # Ne need to force the check pids here, temporary
        self.provider._checkPids = True

        t = pwgui.tree.BoundTree(parent, self.provider, style=LIST_TREEVIEW)
        self.provider._checkPids = False

        t.itemDoubleClick = self._runItemDoubleClick
        t.itemClick = self._runTreeItemClick

        return t

    def updateRunsTree(self, refresh=False):
        self.provider.setRefresh(refresh)
        self.runsTree.update()
        self.updateRunsTreeSelection()

    def updateRunsTreeSelection(self):
        for prot in self._iterSelectedProtocols():
            treeId = self.provider.getObjectFromId(prot.getObjId())._treeId
            self.runsTree.selection_add(treeId)

    def createRunsGraph(self, parent):
        self.runsGraphCanvas = pwgui.Canvas(parent, width=400, height=400,
                                            tooltipCallback=self._runItemTooltip,
                                            tooltipDelay=1000,
                                            name=ProtocolsView.RUNS_CANVAS_NAME,
                                            takefocus=True,
                                            highlightthickness=0)

        self.runsGraphCanvas.onClickCallback = self._runItemClick
        self.runsGraphCanvas.onDoubleClickCallback = self._runItemDoubleClick
        self.runsGraphCanvas.onRightClickCallback = self._runItemRightClick
        self.runsGraphCanvas.onControlClickCallback = self._runItemControlClick
        self.runsGraphCanvas.onAreaSelected = self._selectItemsWithinArea
        self.runsGraphCanvas.onMiddleMouseClickCallback = self._runItemMiddleClick

        parent.grid_columnconfigure(0, weight=1)
        parent.grid_rowconfigure(0, weight=1)

        self.settings.getNodes().updateDict()
        self.settings.getLabels().updateDict()

        self.updateRunsGraph()

    def updateRunsGraph(self, refresh=False, reorganize=False, checkPids=False):

        self.runsGraph = self.project.getRunsGraph(refresh=refresh,
                                                   checkPids=checkPids)
        self.drawRunsGraph(reorganize)

    def drawRunsGraph(self, reorganize=False):

        # Check if there are positions stored
        if reorganize:
            # Create layout to arrange nodes as a level tree
            layout = pwgui.LevelTreeLayout()
            self.runsGraphCanvas.reorganizeGraph(self.runsGraph, layout)
        else:
            self.runsGraphCanvas.clear()

            layout = pwgui.LevelTreeLayout() if len(
                self.settings.getNodes()) == 0 else pwgui.BasicLayout()

            # Create empty nodeInfo for new runs
            for node in self.runsGraph.getNodes():
                nodeId = node.run.getObjId() if node.run else 0
                nodeInfo = self.settings.getNodeById(nodeId)
                if nodeInfo is None:
                    self.settings.addNode(nodeId, x=0, y=0, expanded=True,
                                          visible=True)

            self.runsGraphCanvas.drawGraph(self.runsGraph, layout,
                                           drawNode=self.createRunItem,
                                           nodeList=self.settings.nodeList)

    def createRunItem(self, canvas, node):

        nodeId = node.run.getObjId() if node.run else 0
        nodeInfo = self.settings.getNodeById(nodeId)

        # Extend attributes: use some from nodeInfo
        node.expanded = nodeInfo.isExpanded()
        node.x, node.y = nodeInfo.getPosition()
        node.visible = nodeInfo.isVisible()
        nodeText = self._getNodeText(node)

        # Get status color
        statusColor = getStatusColorFromNode(node)

        # Get the box color (depends on color mode: label or status)
        boxColor = self._getBoxColor(nodeInfo, statusColor, node)

        # Draw the box
        item = RunBox(nodeInfo, self.runsGraphCanvas,
                      nodeText, node.x, node.y,
                      bgColor=boxColor, textColor='black')
        # No border
        item.margin = 0

        # Paint the oval..if apply.
        self._paintOval(item, statusColor)

        # Paint the bottom line (for now only labels are painted there).
        self._paintBottomLine(item)

        item.setSelected(nodeId in self._selection)
        return item

    def _getBoxColor(self, nodeInfo, statusColor, node):

        try:

            # If the color has to go to the box
            if self.settings.statusColorMode():
                boxColor = statusColor

            elif self.settings.ageColorMode():

                if node.run:

                    # Project elapsed time
                    elapsedTime = node.run.getProject().getElapsedTime()
                    creationTime = node.run.getProject().getCreationTime()

                    # Get the latest activity timestamp
                    ts = node.run.endTime.datetime()

                    if elapsedTime is None or creationTime is None or ts is None:
                        boxColor = DEFAULT_BOX_COLOR

                    else:

                        # tc closer to the end are younger
                        protAge = ts - creationTime

                        boxColor = self._ageColor('#6666ff', elapsedTime,
                                                  protAge)
                else:
                    boxColor = DEFAULT_BOX_COLOR

            elif self.settings.sizeColorMode():

                # Get the protocol size
                protSize = self._getRunSize(node)

                boxColor = self._sizeColor(protSize)

            # ... box is for the labels.
            elif self.settings.labelsColorMode():
                # If there is only one label use the box for the color.
                if self._getLabelsCount(nodeInfo) == 1:

                    labelId = nodeInfo.getLabels()[0]
                    label = self.settings.getLabels().getLabel(labelId)

                    # If there is no label (it has been deleted)
                    if label is None:
                        nodeInfo.getLabels().remove(labelId)
                        boxColor = DEFAULT_BOX_COLOR
                    else:
                        boxColor = label.getColor()

                else:
                    boxColor = DEFAULT_BOX_COLOR
            else:
                boxColor = DEFAULT_BOX_COLOR

            return boxColor
        except Exception as e:
            logger.debug("Can't get color for %s. %s" % (node, e))
            return DEFAULT_BOX_COLOR

    @staticmethod
    def _getRunSize(node):
        """
        Returns the size "recursively" of a run

        :param node: node of the graph.
        :return: size in bytes

        """

        if not node.run:
            return 0
        else:
            return node.run.getSize()

    @classmethod
    def _sizeColor(cls, size):
        """
        Returns the color that corresponds to the size
        :param size:
        :return:
        """

        for threshold, color in cls.SIZE_COLORS.items():
            if size <= threshold:
                return color

        return "#000000"
    @staticmethod
    def _ageColor(rgbColorStr, projectAge, protocolAge):

        #  Get the ratio
        ratio = protocolAge.seconds / float(projectAge.seconds)

        # Invert direction: older = white = 100%, newest = rgbColor = 0%
        ratio = 1 - ratio

        # There are cases coming with protocols older than the project.
        ratio = 0 if ratio < 0 else ratio

        hexTuple = pwutils.hex_to_rgb(rgbColorStr)
        lighterTuple = pwutils.lighter(hexTuple, ratio)
        return pwutils.rgb_to_hex(lighterTuple)

    @staticmethod
    def _getLabelsCount(nodeInfo):

        return 0 if nodeInfo.getLabels() is None else len(nodeInfo.getLabels())

    def _paintBottomLine(self, item):

        if self.settings.labelsColorMode():
            self._addLabels(item)

    def _paintOval(self, item, statusColor):
        # Show the status as a circle in the top right corner
        if not self.settings.statusColorMode():
            # Option: Status item.
            (topLeftX, topLeftY, bottomRightX,
             bottomRightY) = self.runsGraphCanvas.bbox(item.id)
            statusSize = 10
            statusX = bottomRightX - (statusSize + 3)
            statusY = topLeftY + 3

            pwgui.Oval(self.runsGraphCanvas, statusX, statusY, statusSize,
                       color=statusColor, anchor=item)

        # in statusColorMode
        else:
            # Show a black circle if there is any label
            if self._getLabelsCount(item.nodeInfo) > 0:
                (topLeftX, topLeftY, bottomRightX,
                 bottomRightY) = self.runsGraphCanvas.bbox(item.id)
                statusSize = 10
                statusX = bottomRightX - (statusSize + 3)
                statusY = topLeftY + 3

                pwgui.Oval(self.runsGraphCanvas, statusX, statusY, statusSize,
                           color='black', anchor=item)

    def _getNodeText(self, node):
        nodeText = node.getLabel()
        # Truncate text to prevent overflow
        if len(nodeText) > 40:
            nodeText = nodeText[:37] + "..."

        if node.run:
            expandedStr = '' if node.expanded else '\n ➕ %s more' % str(node.countChilds({}))
            if self.runsView == VIEW_TREE_SMALL:
                nodeText = node.getName() + expandedStr
            else:
                nodeText += expandedStr + '\n' + node.run.getStatusMessage() if not expandedStr else expandedStr
                if node.run.summaryWarnings:
                    nodeText += u' \u26a0'
        return nodeText

    def _addLabels(self, item):
        # If there is only one label it should be already used in the box color.
        if self._getLabelsCount(item.nodeInfo) < 2:
            return
        # Get the positions of the box
        (topLeftX, topLeftY, bottomRightX,
         bottomRightY) = self.runsGraphCanvas.bbox(item.id)

        # Get the width of the box
        boxWidth = bottomRightX - topLeftX

        # Set the size
        marginV = 3
        marginH = 2
        labelWidth = (boxWidth - (2 * marginH)) / len(item.nodeInfo.getLabels())
        labelHeight = 6

        # Leave some margin on the right and bottom
        labelX = bottomRightX - marginH
        labelY = bottomRightY - (labelHeight + marginV)

        for index, labelId in enumerate(item.nodeInfo.getLabels()):

            # Get the label
            label = self.settings.getLabels().getLabel(labelId)

            # If not none
            if label is not None:
                # Move X one label to the left
                if index == len(item.nodeInfo.getLabels()) - 1:
                    labelX = topLeftX + marginH
                else:
                    labelX -= labelWidth

                pwgui.Rectangle(self.runsGraphCanvas, labelX, labelY,
                                labelWidth, labelHeight, color=label.getColor(),
                                anchor=item)
            else:

                item.nodeInfo.getLabels().remove(labelId)

    def switchRunsView(self):
        viewValue = self.switchCombo.getValue()
        self.runsView = viewValue
        self.settings.setRunsView(viewValue)

        if viewValue == VIEW_LIST:
            self.runsTree.grid(row=0, column=0, sticky='news')
            self.runsGraphCanvas.frame.grid_remove()
            self.updateRunsTree()
            self.viewButtons[ACTION_TREE].grid_remove()
            self._lastRightClickPos = None
        else:
            self.runsTree.grid_remove()
            self.updateRunsGraph()
            self.runsGraphCanvas.frame.grid(row=0, column=0, sticky='news')
            self.viewButtons[ACTION_TREE].grid(row=0, column=1)

    def _protocolItemClick(self, e=None):
        # Get the tree widget that originated the event
        # it could be the left panel protocols tree or just
        # the search protocol dialog tree
        tree = e.widget
        protClassName = tree.getFirst().split('.')[-1]
        protClass = self.domain.getProtocols().get(protClassName)
        prot = self.project.newProtocol(protClass)
        self._openProtocolForm(prot, disableRunMode=True)

    def _toggleColorScheme(self, e=None):

        currentMode = self.settings.getColorMode()

        if currentMode >= len(self.settings.COLOR_MODES) - 1:
            currentMode = -1

        nextColorMode = currentMode + 1

        self.settings.setColorMode(nextColorMode)
        self._updateActionToolbar()
        # self.updateRunsGraph()
        self.drawRunsGraph()
        self._infoAboutColorScheme()

    def _infoAboutColorScheme(self):
        """ Writes in the info widget a brief description abot the color scheme."""

        colorScheme = self.settings.getColorMode()

        msg = "Color mode changed to %s. %s"
        if colorScheme == self.settings.COLOR_MODE_AGE:
            msg = msg % ("AGE", "Young boxes will have an darker color.")
        elif colorScheme == self.settings.COLOR_MODE_SIZE:
            keys = list(self.SIZE_COLORS.keys())
            msg = msg % ("SIZE", "Semaphore color scheme. Green <= %s, Orange <=%s, Red <=%s, Dark quite big." %
                         (pwutils.prettySize(keys[0]),
                          pwutils.prettySize(keys[1]),
                          pwutils.prettySize(keys[2]))
                         )
        elif colorScheme == self.settings.COLOR_MODE_STATUS:
            msg = msg % ("STATUS", "Color based on the status. A black circle indicates it has labels")
        elif colorScheme == self.settings.COLOR_MODE_LABELS:
            msg = msg % ("LABELS", "Color based on custom labels you've assigned. Small circles reflect the protocol status")

        self.info(msg)
    def _toggleDebug(self, e=None):
        Config.toggleDebug()

    def _selectAllProtocols(self, e=None):
        self._selection.clear()

        # WHY GOING TO THE db?
        #  Let's try using in memory data.
        # for prot in self.project.getRuns():
        for prot in self.project.runs:
            self._selection.append(prot.getObjId())
        self._updateSelection()

        # self.updateRunsGraph()
        self.drawRunsGraph()

    def _inspectProtocols(self, e=None):
        objs = self._getSelectedProtocols()
        # We will inspect the selected objects or
        #   the whole project is no protocol is selected
        if len(objs) > 0:
            objs.sort(key=lambda obj: obj._objId, reverse=True)
            filePath = objs[0]._getLogsPath('inspector.csv')
            doInspect = True
        else:
            proj = self.project
            filePath = proj.getLogPath('inspector.csv')
            objs = [proj]
            doInspect = pwgui.dialog.askYesNo(Message.TITLE_INSPECTOR,
                                              Message.LABEL_INSPECTOR, self.root)

        if doInspect:
            inspectObj(objs, filePath)
            # we open the resulting CSV file with the OS default software
            pwgui.text.openTextFileEditor(filePath)

    # NOt used!: pconesa 02/11/2016.
    # def _deleteSelectedProtocols(self, e=None):
    #
    #     for selection in self._selection:
    #         self.project.getProtocol(self._selection[0])
    #
    #
    #     self._updateSelection()
    #     self.updateRunsGraph()

    def _updateSelection(self):
        self._fillSummary()
        self._fillMethod()
        self._fillLogs()
        self._showHideAnalyzeResult()

        if self._isSingleSelection():
            last = self.getSelectedProtocol()
            self._lastSelectedProtId = last.getObjId() if last else None

        self._updateActionToolbar()

    def _runTreeItemClick(self, item=None):
        self._selection.clear()
        for prot in self.runsTree.iterSelectedObjects():
            self._selection.append(prot.getObjId())
        self._updateSelection()

    def _selectItemProtocol(self, prot):
        """ Call this function when a new box (item) of a protocol
        is selected. It should be called either from itemClick
        or itemRightClick
        """
        self._selection.clear()
        self.settings.dataSelection.clear()
        self._selection.append(prot.getObjId())

        # Select output data too
        self.toggleDataSelection(prot, True)

        self._updateSelection()
        self.runsGraphCanvas.update_idletasks()

    def _deselectItems(self, item):
        """ Deselect all items except the item one
        """
        g = self.project.getRunsGraph()

        for node in g.getNodes():
            if node.run and node.run.getObjId() in self._selection:
                # This option is only for compatibility with all projects
                if hasattr(node, 'item'):
                    node.item.setSelected(False)
        item.setSelected(True)

    def _runItemClick(self, item=None):

        # If click is in a empty area....start panning
        if item is None:
            print("Click on empty area")
            return

        self.runsGraphCanvas.focus_set()

        # Get last selected item for tree or graph
        if self.runsView == VIEW_LIST:
            prot = self.project.mapper.selectById(int(self.runsTree.getFirst()))
        else:
            prot = item.node.run
            if prot is None:  # in case it is the main "Project" node
                return
            self._deselectItems(item)
        self._selectItemProtocol(prot)

    def _runItemDoubleClick(self, e=None):
        if e.nodeInfo.isExpanded():
            self._runActionClicked(ACTION_EDIT)

    def _runItemMiddleClick(self, e=None):
        self._runActionClicked(ACTION_SELECT_TO)

    def _runItemRightClick(self, item=None):
        prot = item.node.run
        if prot is None:  # in case it is the main "Project" node
            return
        n = len(self._selection)
        # Only select item with right-click if there is a single
        # item selection, not for multiple selection
        if n <= 1:
            self._deselectItems(item)
            self._selectItemProtocol(prot)
            self._lastRightClickPos = self.runsGraphCanvas.eventPos

        return self.provider.getObjectActions(prot)

    def _runItemControlClick(self, item=None):
        # Get last selected item for tree or graph
        if self.runsView == VIEW_LIST:
            # TODO: Prot is not used!!
            prot = self.project.mapper.selectById(int(self.runsTree.getFirst()))
        else:
            prot = item.node.run
            protId = prot.getObjId()
            if protId in self._selection:
                item.setSelected(False)
                self._selection.remove(protId)

                # Remove data selected
                self.toggleDataSelection(prot, False)
            else:

                item.setSelected(True)
                if len(self._selection) == 1:  # repaint first selected item
                    firstSelectedNode = self.runsGraph.getNode(str(self._selection[0]))
                    if hasattr(firstSelectedNode, 'item'):
                        firstSelectedNode.item.setSelected(False)
                        firstSelectedNode.item.setSelected(True)
                self._selection.append(prot.getObjId())

                # Select output data too
                self.toggleDataSelection(prot, True)

        self._updateSelection()

    def toggleDataSelection(self, prot, append):

        # Go through the data selection
        for paramName, output in prot.iterOutputAttributes():
            if append:
                self.settings.dataSelection.append(output.getObjId())
            else:
                self.settings.dataSelection.remove(output.getObjId())

    def _runItemTooltip(self, tw, item):
        """ Create the contents of the tooltip to be displayed
        for the given item.
        Params:
            tw: a tk.TopLevel instance (ToolTipWindow)
            item: the selected item.
        """
        prot = item.node.run

        if prot:
            tm = '*%s*\n' % prot.getRunName()
            tm += 'Identifier :%s\n' % prot.getObjId()
            tm += 'Status: %s\n' % prot.getStatusMessage()
            tm += 'Wall time: %s\n' % pwutils.prettyDelta(prot.getElapsedTime())
            tm += 'CPU time: %s\n' % pwutils.prettyDelta(dt.timedelta(seconds=prot.cpuTime))
            tm += 'Folder size: %s\n' % pwutils.prettySize(prot.getSize())

            if not hasattr(tw, 'tooltipText'):
                frame = tk.Frame(tw)
                frame.grid(row=0, column=0)
                tw.tooltipText = pwgui.dialog.createMessageBody(frame, tm, None,
                                                                textPad=0,
                                                                textBg=Color.LIGHT_GREY_COLOR_2)
                tw.tooltipText.config(bd=1, relief=tk.RAISED)
            else:
                pwgui.dialog.fillMessageText(tw.tooltipText, tm)

    @staticmethod
    def _selectItemsWithinArea(x1, y1, x2, y2, enclosed=False):
        """
        Parameters
        ----------
        x1: x coordinate of first corner of the area
        y1: y coordinate of first corner of the area
        x2: x coordinate of second corner of the area
        y2: y coordinate of second corner of the area
        enclosed: Default True. Returns enclosed items,
                  overlapping items otherwise.
        Returns
        -------
        Nothing

        """

        return
        # NOT working properly: Commented for the moment.
        # if enclosed:
        #     items = self.runsGraphCanvas.find_enclosed(x1, y1, x2, y2)
        # else:
        #     items = self.runsGraphCanvas.find_overlapping(x1, y1, x2, y2)
        #
        # update = False
        #
        # for itemId in items:
        #     if itemId in self.runsGraphCanvas.items:
        #
        #         item = self.runsGraphCanvas.items[itemId]
        #         if not item.node.isRoot():
        #             item.setSelected(True)
        #             self._selection.append(itemId)
        #             update = True
        #
        # if update is not None: self._updateSelection()

    def _openProtocolForm(self, prot, disableRunMode=False):
        """Open the Protocol GUI Form given a Protocol instance"""

        w = FormWindow(Message.TITLE_NAME_RUN + prot.getClassName(),
                       prot, self._executeSaveProtocol, self.windows,
                       hostList=self.project.getHostNames(),
                       updateProtocolCallback=self._updateProtocol,
                       disableRunMode=disableRunMode)
        w.adjustSize()
        w.show(center=True)

    def _browseSteps(self):
        """ Open a new window with the steps list. """
        window = StepsWindow(Message.TITLE_BROWSE_DATA, self.windows,
                             self.getSelectedProtocol())
        window.show()

    def _browseRunData(self):
        provider = ProtocolTreeProvider(self.getSelectedProtocol())
        window = pwgui.browser.BrowserWindow(Message.TITLE_BROWSE_DATA,
                                             self.windows)
        window.setBrowser(pwgui.browser.ObjectBrowser(window.root, provider))
        window.itemConfig(self.getSelectedProtocol(), open=True)
        window.show()

    def _browseRunDirectory(self):
        """ Open a file browser to inspect the files generated by the run. """
        protocol = self.getSelectedProtocol()
        workingDir = protocol.getWorkingDir()
        if os.path.exists(workingDir):

            protFolderShortCut = ShortCut.factory(workingDir,name="Protocol folder", icon=None ,toolTip="Protocol directory")
            window = pwgui.browser.FileBrowserWindow("Browsing: " + workingDir,
                                                     master=self.windows,
                                                     path=workingDir,
                                                     shortCuts=[protFolderShortCut])
            window.show()
        else:
            self.windows.showInfo("Protocol working dir does not exists: \n %s"
                                  % workingDir)

    def _iterSelectedProtocols(self):
        for protId in sorted(self._selection):
            prot = self.project.getRunsGraph().getNode(str(protId)).run
            if prot:
                yield prot

    def _getSelectedProtocols(self):
        return [prot for prot in self._iterSelectedProtocols()]

    def _iterSelectedNodes(self):

        for protId in sorted(self._selection):
            node = self.settings.getNodeById(protId)

            yield node

    def _getSelectedNodes(self):
        return [node for node in self._iterSelectedNodes()]

    def getSelectedProtocol(self):
        if self._selection:
            return self.project.getProtocol(self._selection[0])
        return None

    def _showHideAnalyzeResult(self):

        if self._selection:
            self.btnAnalyze.grid()
        else:
            self.btnAnalyze.grid_remove()

    def _fillSummary(self):
        self.summaryText.setReadOnly(False)
        self.summaryText.clear()
        self.infoTree.clear()
        n = len(self._selection)

        if n == 1:
            prot = self.getSelectedProtocol()

            if prot:
                provider = RunIOTreeProvider(self, prot, self.project.mapper, self.info)
                self.infoTree.setProvider(provider)
                self.infoTree.grid(row=0, column=0, sticky='news')
                self.infoTree.update_idletasks()
                # Update summary
                self.summaryText.addText(prot.summary())
            else:
                self.infoTree.clear()

        elif n > 1:
            self.infoTree.clear()
            for prot in self._iterSelectedProtocols():
                self.summaryText.addLine('> _%s_' % prot.getRunName())
                for line in prot.summary():
                    self.summaryText.addLine(line)
                self.summaryText.addLine('')
        self.summaryText.setReadOnly(True)

    def _fillMethod(self):

        try:
            self.methodText.setReadOnly(False)
            self.methodText.clear()
            self.methodText.addLine("*METHODS:*")
            cites = OrderedDict()

            for prot in self._iterSelectedProtocols():
                self.methodText.addLine('> _%s_' % prot.getRunName())
                for line in prot.getParsedMethods():
                    self.methodText.addLine(line)
                cites.update(prot.getCitations())
                cites.update(prot.getPackageCitations())
                self.methodText.addLine('')

            if cites:
                self.methodText.addLine('*REFERENCES:*   '
                                        ' [[sci-bib:][<<< Open as bibtex >>>]]')
                for cite in cites.values():
                    self.methodText.addLine(cite)

            self.methodText.setReadOnly(True)
        except Exception as e:
            self.methodText.addLine('Could not load all methods:' + str(e))

    def _fillLogs(self):
        try:
            prot = self.getSelectedProtocol()

            if not self._isSingleSelection() or not prot:
                self.outputViewer.clear()
                self._lastStatus = None
            elif prot.getObjId() != self._lastSelectedProtId:
                self._lastStatus = prot.getStatus()
                i = self.outputViewer.getIndex()
                self.outputViewer.clear()
                # Right now skip the err tab since we are redirecting
                # stderr to stdout
                out, err, schedule = prot.getLogPaths()
                self.outputViewer.addFile(out)
                self.outputViewer.addFile(err)
                if os.path.exists(schedule):
                    self.outputViewer.addFile(schedule)
                elif i == 2:
                    i = 0
                self.outputViewer.setIndex(i)  # Preserve the last selected tab
                self.outputViewer.selectedText().goEnd()
                # when there are not logs, force re-load next time
                if (not os.path.exists(out) or
                        not os.path.exists(err)):
                    self._lastStatus = None

            elif prot.isActive() or prot.getStatus() != self._lastStatus:
                doClear = self._lastStatus is None
                self._lastStatus = prot.getStatus()
                self.outputViewer.refreshAll(clear=doClear, goEnd=doClear)
        except Exception as e:
            self.info("Something went wrong filling %s's logs: %s. Check terminal for details" % (prot, e))
            import traceback
            traceback.print_exc()

    def _scheduleRunsUpdate(self, secs=1):
        # self.runsTree.after(secs*1000, self.refreshRuns)
        self.windows.enqueue(self.refreshRuns)

    def executeProtocol(self, prot):
        """ Function to execute a protocol called not
        directly from the Form "Execute" button.
        """
        # We need to equeue the execute action
        # to be executed in the same thread
        self.windows.enqueue(lambda: self._executeSaveProtocol(prot))

    def _executeSaveProtocol(self, prot, onlySave=False, doSchedule=False):
        if onlySave:
            self.project.saveProtocol(prot)
            msg = Message.LABEL_SAVED_FORM
            # msg = "Protocol successfully saved."

        else:
            if doSchedule:
                self.project.scheduleProtocol(prot)
            else:
                self.project.launchProtocol(prot)
            # Select the launched protocol to display its summary, methods..etc
            self._selection.clear()
            self._selection.append(prot.getObjId())
            self._updateSelection()
            self._lastStatus = None  # clear lastStatus to force re-load the logs
            msg = ""

        # Update runs list display, even in save we
        # need to get the updated copy of the protocol
        self._scheduleRunsUpdate()
        self._selectItemProtocol(prot)

        return msg

    def _updateProtocol(self, prot):
        """ Callback to notify about the change of a protocol
        label or comment. 
        """
        self._scheduleRunsUpdate()

    def _continueProtocol(self, prot):
        self.project.continueProtocol(prot)
        self._scheduleRunsUpdate()

    def _onDelPressed(self):
        # This function will be connected to the key 'Del' press event
        # We need to check if the canvas have the focus and then
        # proceed with the delete action

        # get the widget with the focus
        widget = self.focus_get()

        # Call the delete action only if the widget is the canvas
        if str(widget).endswith(ProtocolsView.RUNS_CANVAS_NAME):
            try:
                self._deleteProtocol()
            except Exception as ex:
                self.windows.showError(str(ex))

    def _deleteProtocol(self):
        protocols = self._getSelectedProtocols()

        if len(protocols) == 0:
            return

        protStr = '\n  - '.join(['*%s*' % p.getRunName() for p in protocols])

        if pwgui.dialog.askYesNo(Message.TITLE_DELETE_FORM,
                                 Message.LABEL_DELETE_FORM % protStr,
                                 self.root):
            self.info('Deleting protocols...')
            self.project.deleteProtocol(*protocols)
            self._selection.clear()
            self._updateSelection()
            self._scheduleRunsUpdate()
            self.cleanInfo()

    def _F2Pressed(self, event):
        """ Invoked then F2 if pressed: Protocol rename"""
        self._runActionClicked(ACTION_RENAME)

    def _editProtocol(self, protocol):
        disableRunMode = False
        if protocol.isSaved():
            disableRunMode = True
        self._openProtocolForm(protocol, disableRunMode=disableRunMode)

    def _copyProtocols(self):
        protocols = self._getSelectedProtocols()
        if len(protocols) == 1:
            newProt = self.project.copyProtocol(protocols[0])
            if newProt is None:
                self.windows.showError("Error copying protocol.!!!")
            else:
                self._openProtocolForm(newProt, disableRunMode=True)
        else:
            self.info('Copying the protocols...')
            self.project.copyProtocol(protocols)
            self.refreshRuns()
            self.cleanInfo()

    def _stopWorkFlow(self, action):

        protocols = self._getSelectedProtocols()
        workflowProtocolList, activeProtList = self.project._getWorkflowFromProtocol(protocols[0],
                                                                                     False)
        if activeProtList:
            errorProtList = []
            if pwgui.dialog.askYesNo(Message.TITLE_STOP_WORKFLOW_FORM,
                                     Message.TITLE_STOP_WORKFLOW, self.root):
                self.info('Stopping the workflow...')
                errorProtList = self.project.stopWorkFlow(activeProtList)
                self.cleanInfo()
                self.refreshRuns()
            if errorProtList:
                msg = '\n'
                for prot in errorProtList:
                    msg += str(prot.getObjLabel()) + '\n'
                pwgui.dialog.MessageDialog(
                    self, Message.TITLE_STOPPED_WORKFLOW_FAILED,
                    Message.TITLE_STOPPED_WORKFLOW_FAILED + ' with: ' + msg,
                    Icon.ERROR)

    def _resetWorkFlow(self, action):

        protocols = self._getSelectedProtocols()
        errorProtList = []
        if pwgui.dialog.askYesNo(Message.TITLE_RESET_WORKFLOW_FORM,
                                 Message.TITLE_RESET_WORKFLOW, self.root):
            self.info('Resetting the workflow...')
            workflowProtocolList, activeProtList = self.project._getWorkflowFromProtocol(protocols[0])
            errorProtList = self.project.resetWorkFlow(workflowProtocolList)
            self.cleanInfo()
            self.refreshRuns()
        if errorProtList:
            msg = '\n'
            for prot in errorProtList:
                msg += str(prot.getObjLabel()) + '\n'
            pwgui.dialog.MessageDialog(
                self, Message.TITLE_RESETED_WORKFLOW_FAILED,
                Message.TITLE_RESETED_WORKFLOW_FAILED + ' with: ' + msg,
                Icon.ERROR)

    def _launchWorkFlow(self, action):
        """
        This function can launch a workflow from a selected protocol in two
        modes depending on the 'action' value (RESTART, CONTINUE)
        """
        protocols = self._getSelectedProtocols()
        errorList = []
        defaultMode = pwprot.MODE_RESUME
        workflowProtocolList, activeProtList = self.project._getWorkflowFromProtocol(protocols[0])

        # Check if exists active protocols
        if activeProtList:
            msg = '\n'
            for activeProt in activeProtList:
                msg += str(activeProt.getObjLabel()) + '\n'
            pwgui.dialog.MessageDialog(
                self, Message.TITLE_LAUNCHED_WORKFLOW_FAILED_FORM,
                Message.TITLE_LAUNCHED_WORKFLOW_FAILED + "\n" +
                Message.TITLE_ACTIVE_PROTOCOLS + "\n" + msg,
                Icon.ERROR)

        elif action == ACTION_RESTART_WORKFLOW:
            if pwgui.dialog.askYesNo(Message.TITLE_RESTART_WORKFLOW_FORM,
                                     Message.TITLE_RESTART_WORKFLOW, self.root):
                defaultMode = pwprot.MODE_RESTART
                self.info('Restarting...')
                errorList = self.project.launchWorkflow(workflowProtocolList,
                                                        defaultMode)
                self.cleanInfo()
        elif action == ACTION_CONTINUE_WORKFLOW:
            self.info('Continuing...')
            errorList = self.project.launchWorkflow(workflowProtocolList,
                                                    defaultMode)
            self.cleanInfo()

        if errorList:
            msg = ''
            for errorProt in errorList:
                msg += str(errorProt) + '\n'
            pwgui.dialog.MessageDialog(
                self, Message.TITLE_LAUNCHED_WORKFLOW_FAILED_FORM,
                Message.TITLE_LAUNCHED_WORKFLOW_FAILED + "\n" + msg,
                Icon.ERROR)
        self.refreshRuns()

    def _selectLabels(self):
        selectedNodes = self._getSelectedNodes()

        if selectedNodes:
            dlg = self.windows.manageLabels()

            if dlg.resultYes():
                for node in selectedNodes:
                    node.setLabels([label.getName() for label in dlg.values])

                # self.updateRunsGraph()
                self.drawRunsGraph()

    def _selectAncestors(self):
        self._selectNodes(down=False)

    def _selectDescendants(self):
        self._selectNodes(down=True)

    def _selectNodes(self, down=True, fromRun=None):
        """ Selects all nodes in the specified direction, defaults to down."""
        nodesToSelect = []
        # If parent param not passed...
        if fromRun is None:
            # ..use selection, must be first call
            for protId in self._selection:
                run = self.runsGraph.getNode(str(protId))
                nodesToSelect.append(run)
        else:
            name = fromRun.getName()

            if not name.isdigit():
                return
            else:
                name = int(name)

            # If already selected (may be this should be centralized)
            if name not in self._selection:
                nodesToSelect = (fromRun,)
                self._selection.append(name)

        # Go in the direction .
        for run in nodesToSelect:
            # Choose the direction: down or up.
            direction = run.getChilds if down else run.getParents

            # Select himself plus ancestors
            for parent in direction():
                self._selectNodes(down, parent)

        # Only update selection at the end, avoid recursion
        if fromRun is None:
            self._lastSelectedProtId = None
            self._updateSelection()
            self.drawRunsGraph()


    def _exportProtocols(self, defaultPath=None, defaultBasename=None):
        protocols = self._getSelectedProtocols()

        def _export(obj):
            filename = os.path.join(browser.getCurrentDir(),
                                    browser.getEntryValue())
            try:
                if (not os.path.exists(filename) or
                    self.windows.askYesNo("File already exists",
                                          "*%s* already exists, do you want "
                                          "to overwrite it?" % filename)):
                    self.project.exportProtocols(protocols, filename)
                    logger.info("Workflow successfully saved to '%s' "
                                          % filename)
                else:  # try again
                    self._exportProtocols(defaultPath=browser.getCurrentDir(),
                                          defaultBasename=browser.getEntryValue())
            except Exception as ex:
                import traceback
                traceback.print_exc()
                self.windows.showError(str(ex))

        browser = pwgui.browser.FileBrowserWindow(
            "Choose .json file to save workflow",
            master=self.windows,
            path=defaultPath or self.project.getPath(''),
            onSelect=_export,
            entryLabel='File  ', entryValue=defaultBasename or 'workflow.json')
        browser.show()

    def _exportUploadProtocols(self):
        try:
            jsonFn = os.path.join(tempfile.mkdtemp(), 'workflow.json')
            self.project.exportProtocols(self._getSelectedProtocols(), jsonFn)
            WorkflowRepository().upload(jsonFn)
            pwutils.cleanPath(jsonFn)
        except Exception as ex:
            self.windows.showError("Error connecting to workflow repository:\n"
                                   + str(ex))

    def _stopProtocol(self, prot):
        if pwgui.dialog.askYesNo(Message.TITLE_STOP_FORM,
                                 Message.LABEL_STOP_FORM, self.root):
            self.project.stopProtocol(prot)
            self._lastStatus = None  # force logs to re-load
            self._scheduleRunsUpdate()

    def _analyzeResults(self, prot):
        viewers = self.domain.findViewers(prot.getClassName(), DESKTOP_TKINTER)
        if len(viewers):
            # Instantiate the first available viewer
            # TODO: If there are more than one viewer we should display
            # TODO: a selection menu
            firstViewer = viewers[0](project=self.project, protocol=prot,
                                     parent=self.windows)

            if isinstance(firstViewer, ProtocolViewer):
                firstViewer.visualize(prot, windows=self.windows)
            else:
                firstViewer.visualize(prot)
        else:
            outputList = []
            for _, output in prot.iterOutputAttributes():
                outputList.append(output)

            for output in outputList:
                viewers = self.domain.findViewers(output.getClassName(), DESKTOP_TKINTER)
                if len(viewers):
                    # Instantiate the first available viewer
                    # TODO: If there are more than one viewer we should display
                    # TODO: a selection menu
                    viewerclass = viewers[0]
                    firstViewer = viewerclass(project=self.project,
                                              protocol=prot,
                                              parent=self.windows)
                    # FIXME:Probably o longer needed protocol on args, already provided on init
                    firstViewer.visualize(output, windows=self.windows,
                                          protocol=prot)

    def _analyzeResultsClicked(self, e=None):
        """ Function called when button "Analyze results" is called. """
        prot = self.getSelectedProtocol()

        # Nothing selected
        if prot is None:
            return

        if os.path.exists(prot._getPath()):
            self._analyzeResults(prot)
        else:
            self.windows.showInfo("Selected protocol hasn't been run yet.")

    def _bibExportClicked(self, e=None):
        try:
            bibTexCites = OrderedDict()
            for prot in self._iterSelectedProtocols():
                bibTexCites.update(prot.getCitations(bibTexOutput=True))
                bibTexCites.update(prot.getPackageCitations(bibTexOutput=True))

            if bibTexCites:
                with tempfile.NamedTemporaryFile(suffix='.bib') as bibFile:
                    for refId, refDict in bibTexCites.items():
                        # getCitations does not always return a dictionary
                        # if the citation is not found in the bibtex file it adds just
                        # the refId: like "Ramirez-Aportela-2019"
                        # we need to exclude this
                        if isinstance(refDict, dict):
                            refType = refDict['ENTRYTYPE']
                            # remove 'type' and 'id' keys
                            refDict = {k: v for k, v in refDict.items()
                                       if k not in ['ENTRYTYPE', 'ID']}
                            jsonStr = json.dumps(refDict, indent=4,
                                                 ensure_ascii=False)[1:]
                            jsonStr = jsonStr.replace('": "', '"= "')
                            jsonStr = re.sub('(?<!= )"(\S*?)"', '\\1', jsonStr)
                            jsonStr = jsonStr.replace('= "', ' = "')
                            refStr = '@%s{%s,%s\n\n' % (refType, refId, jsonStr)
                            bibFile.write(refStr.encode('utf-8'))
                        else:
                            print("WARNING: reference %s not properly defined or unpublished." % refId)
                    # flush so we can see content when opening
                    bibFile.flush()
                    pwgui.text.openTextFileEditor(bibFile.name)

        except Exception as ex:
            self.windows.showError(str(ex))

        return

    def _renameProtocol(self, prot):
        """ Open the EditObject dialog to edit the protocol name. """
        kwargs = {}
        if self._lastRightClickPos:
            kwargs['position'] = self._lastRightClickPos

        dlg = pwgui.dialog.EditObjectDialog(self.runsGraphCanvas, Message.TITLE_EDIT_OBJECT,
                                            prot, self.project.mapper, **kwargs)
        if dlg.resultYes():
            self._updateProtocol(prot)

    def _runActionClicked(self, action):
        prot = self.getSelectedProtocol()
        if prot:
            try:
                if action == ACTION_DEFAULT:
                    pass
                elif action == ACTION_EDIT:
                    self._editProtocol(prot)
                elif action == ACTION_RENAME:
                    self._renameProtocol(prot)
                elif action == ACTION_COPY:
                    self._copyProtocols()
                elif action == ACTION_DELETE:
                    self._deleteProtocol()
                elif action == ACTION_STEPS:
                    self._browseSteps()
                elif action == ACTION_BROWSE:
                    self._browseRunDirectory()
                elif action == ACTION_DB:
                    self._browseRunData()
                elif action == ACTION_STOP:
                    self._stopProtocol(prot)
                elif action == ACTION_CONTINUE:
                    self._continueProtocol(prot)
                elif action == ACTION_RESULTS:
                    self._analyzeResults(prot)
                elif action == ACTION_EXPORT:
                    self._exportProtocols(defaultPath=pwutils.getHomePath())
                elif action == ACTION_EXPORT_UPLOAD:
                    self._exportUploadProtocols()
                elif action == ACTION_COLLAPSE:
                    node = self.runsGraph.getNode(str(prot.getObjId()))
                    nodeInfo = self.settings.getNodeById(prot.getObjId())
                    nodeInfo.setExpanded(False)
                    self.setVisibleNodes(node, visible=False)
                    self.updateRunsGraph(True, reorganize=False)
                    self._updateActionToolbar()
                elif action == ACTION_EXPAND:
                    node = self.runsGraph.getNode(str(prot.getObjId()))
                    nodeInfo = self.settings.getNodeById(prot.getObjId())
                    nodeInfo.setExpanded(True)
                    self.setVisibleNodes(node, visible=True)
                    self.updateRunsGraph(True, reorganize=False)
                    self._updateActionToolbar()
                elif action == ACTION_LABELS:
                    self._selectLabels()
                elif action == ACTION_SELECT_FROM:
                    self._selectDescendants()
                elif action == ACTION_SELECT_TO:
                    self._selectAncestors()
                elif action == ACTION_RESTART_WORKFLOW:
                    self._launchWorkFlow(action)
                elif action == ACTION_CONTINUE_WORKFLOW:
                    self._launchWorkFlow(action)
                elif action == ACTION_STOP_WORKFLOW:
                    self._stopWorkFlow(action)
                elif action == ACTION_RESET_WORKFLOW:
                    self._resetWorkFlow(action)
                elif action == ACTION_SEARCH:
                    self._searchProtocol()

            except Exception as ex:
                self.windows.showError(str(ex))
                if Config.debugOn():
                    import traceback
                    traceback.print_exc()

        # Following actions do not need a select run
        if action == ACTION_TREE:
            self.drawRunsGraph(reorganize=True)
        elif action == ACTION_REFRESH:
            self.refreshRuns(checkPids=True)

        elif action == ACTION_SWITCH_VIEW:
            self.switchRunsView()

    def setVisibleNodes(self, node, visible=True):
        hasParentHidden = False
        for child in node.getChilds():
            prot = child.run
            nodeInfo = self.settings.getNodeById(prot.getObjId())
            if visible:
                hasParentHidden = self.hasParentHidden(child)
            if not hasParentHidden:
                nodeInfo.setVisible(visible)
                self.setVisibleNodes(child, visible)

    def hasParentHidden(self, node):
        for parent in node.getParents():
            prot = parent.run
            nodeInfo = self.settings.getNodeById(prot.getObjId())
            if not nodeInfo.isVisible() or not nodeInfo.isExpanded():
                return True
        return False


class RunBox(pwgui.TextBox):
    """ Just override TextBox move method to keep track of 
    position changes in the graph.
    """

    def __init__(self, nodeInfo, canvas, text, x, y, bgColor, textColor):
        pwgui.TextBox.__init__(self, canvas, text, x, y, bgColor, textColor)
        self.nodeInfo = nodeInfo
        canvas.addItem(self)

    def move(self, dx, dy):
        pwgui.TextBox.move(self, dx, dy)
        self.nodeInfo.setPosition(self.x, self.y)

    def moveTo(self, x, y):
        pwgui.TextBox.moveTo(self, x, y)
        self.nodeInfo.setPosition(self.x, self.y)






import logging
logger = logging.getLogger(__file__)
import ast
import importlib
import inspect
import json
import os
import shutil
import sys
import types

from .constants import *

HOME = os.path.abspath(os.path.dirname(__file__))
PYTHON = os.environ.get(SCIPION_PYTHON, SCIPION_PYTHON_DEFAULT)


def join(*paths):
    """ join paths from HOME . """
    return os.path.join(HOME, *paths)


__resourcesPath = [join('resources')]


def findResource(filename):
    from .utils.path import findFile
    return findFile(filename, *__resourcesPath)


def genNotesHeading():
    return SCIPION_NOTES_HEADING_MSG


def getAppsPath(*paths):
    return join(APPS, *paths)


def getSyncDataScript():
    return getAppsPath(PW_SYNC_DATA)


def getScheduleScript():
    return getAppsPath(PW_SCHEDULE_RUN)


def getPwProtMpiRunScript():
    return getAppsPath(PW_PROTOCOL_MPIRUN)


def getTestsScript():
    return getAppsPath(PW_RUN_TESTS)


def getViewerScript():
    return getAppsPath(PW_VIEWER)


def getPyworkflowPath():
    """ Returns the path where pyworkflow is"""
    return os.path.dirname(__file__)


def getModuleFolder(moduleName):
    """ Returns the path of a module without importing it"""
    spec = importlib.util.find_spec(moduleName)
    return os.path.dirname(spec.origin)


class Config:
    """ Main Config for pyworkflow. It contains the main Scipion configuration variables
    providing default values or, if present, taking them from the environment.
    Necessary value is SCIPION_HOME and has to be present in the environment"""

    @staticmethod
    def __get(key, default):
        value = os.environ.get(key, default)
        # Expand user and variables if string value
        if isinstance(value, str):
            value = os.path.expandvars(os.path.expanduser(value))

        return value

    class Root:
        """ Simple helper to return path from a root. """

        def __init__(self, root):
            self._root = root

        def join(self, *path):
            # We need to consider variable in the config with ~
            expanded = os.path.expanduser(os.path.join(*path))

            # join will not join if expanded is absolute
            return os.path.join(self._root, expanded)

    # Home for scipion
    _get = __get.__func__
    SCIPION_HOME = os.path.abspath(_get(SCIPION_HOME_VAR, ''))
    "Path where Scipion is installed. Other paths are based on this like SCIPION_SOFTWARE, SCIPION_TESTS,... unless specified"

    _root = Root(SCIPION_HOME)
    _join = _root.join

    CONDA_ACTIVATION_CMD = _get(CONDA_ACTIVATION_CMD_VAR,'')
    "Command to activate/initialize conda itself. Do not confuse it with 'conda activate'. It should be defined at installation time. It looks like this: eval \"$(/extra/miniconda3/bin/conda shell.bash hook)\""

    # SCIPION PATHS
    SCIPION_SOFTWARE = _join(_get('SCIPION_SOFTWARE', 'software'))
    "Path where Scipion will install the software. Defaults to SCIPION_HOME/software."

    SCIPION_TESTS = _join(_get('SCIPION_TESTS', os.path.join('data', 'tests')))
    "Path where to find/download test data. Defaults to SCIPION_HOME/data/tests."

    # User dependent paths
    SCIPION_USER_DATA = _get('SCIPION_USER_DATA', '~/ScipionUserData')
    "Path where Scipion projects are or will be created. Defaults to ~/ScipionUserData"

    SCIPION_TMP = _get('SCIPION_TMP', _join(SCIPION_USER_DATA, 'tmp'))
    "General purpose scipion tmp folder. Defaults to SCIPION_USER_DATA/tmp"

    # LOGGING variables
    SCIPION_LOGS = _get('SCIPION_LOGS', _join(SCIPION_USER_DATA, 'logs'))
    "Path for Scipion logs folder used by the GUI. Defaults to SCIPION_USER_DATA/logs."

    SCIPION_LOG_CONFIG = _get('SCIPION_LOG_CONFIG', None)
    "Optional. Path to a python logging configuration file fine tune the logging."

    SCIPION_LOG = _join(SCIPION_LOGS, 'scipion.log')
    "Path to the file where scipion will write GUI logging messages. Defaults to SCIPION_LOGS/scipion.log"

    SCIPION_LOG_FORMAT = _get('SCIPION_LOG_FORMAT', "%(message)s")
    "Format for all the log lines, defaults to %(message)s. To compose the format see https://docs.python.org/3/library/logging.html#logrecord-attributes"

    SCIPION_LOG_LEVEL = _get(SCIPION_LOG_LEVEL, 'INFO')
    "Default logging level. String among CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET. Default value is INFO."

    NO_COLOR = _get('NO_COLOR', '')
    "Comply with https://no-color.org/ initiative. Set it to something different than '' to deactivate colors in the output."

    SCIPION_SCRATCH = _get(SCIPION_SCRATCH, None)
    "Optional. Path to a location mounted in a scratch drive (SSD,...)"

    SCIPION_TESTS_OUTPUT = _get('SCIPION_TESTS_OUTPUT', _join(SCIPION_USER_DATA, 'Tests'))
    "Path to a folder Where the output of the tests will be written. Defaults to SCIPION_USER_DATA/Tests."

    SCIPION_TEST_NOSYNC = _get('SCIPION_TEST_NOSYNC', "False")
    "Set it to 1, True, Yes or y to cancel test dataset synchronization. Needed when updating files in a dataset."

    SCIPION_SUPPORT_EMAIL = _get('SCIPION_SUPPORT_EMAIL', 'scipion@cnb.csic.es')

    SCIPION_LOGO = _get('SCIPION_LOGO', 'scipion_logo.gif')

    # Config variables
    SCIPION_CONFIG = _get('SCIPION_CONFIG', 'scipion.conf')
    "Path to the scipion configuration file where all this variables could be defined."

    SCIPION_LOCAL_CONFIG = _get('SCIPION_LOCAL_CONFIG', SCIPION_CONFIG)
    "Path to an optional/extra/user configuration file meant to overwrite default variables."

    SCIPION_HOSTS = _get('SCIPION_HOSTS', 'hosts.conf')
    "Path to the host.cof file to allow scipion to use queue engines and run in HPC environments."

    SCIPION_PROTOCOLS = _get('SCIPION_PROTOCOLS', 'protocols.conf')
    ""

    SCIPION_PLUGIN_JSON = _get('SCIPION_PLUGIN_JSON', None)
    "Optional. Path to get the json file with all the plugins available for Scipion."

    SCIPION_PLUGIN_REPO_URL = _get('SCIPION_PLUGIN_REPO_URL',
                                   'http://scipion.i2pc.es/getplugins/')
    "Url from where to get the list of plugins."

    # REMOTE Section
    SCIPION_URL = _get('SCIPION_URL', 'http://scipion.cnb.csic.es/downloads/scipion')
    SCIPION_URL_SOFTWARE = _get('SCIPION_URL_SOFTWARE', SCIPION_URL + '/software')
    SCIPION_URL_TESTDATA = _get('SCIPION_URL_TESTDATA', SCIPION_URL + '/data/tests')

    # Scipion Notes
    SCIPION_NOTES_FILE = _get(SCIPION_NOTES_FILE, 'notes.txt')
    "Name of the file where to write per project notes."

    SCIPION_NOTES_PROGRAM = _get(SCIPION_NOTES_PROGRAM, None)
    "Command or program to use to open the notes file. Otherwise system will extension association will take place."

    SCIPION_NOTES_ARGS = _get(SCIPION_NOTES_ARGS, None)

    # Aspect
    SCIPION_FONT_NAME = _get('SCIPION_FONT_NAME', "Helvetica")
    "Name of the font to use in Scipion GUI. Defaults to Helvetica."

    SCIPION_FONT_SIZE = int(_get('SCIPION_FONT_SIZE', SCIPION_DEFAULT_FONT_SIZE))
    "Size of the 'normal' font to be used in Scipion GUI. Defaults to 10."

    WIZARD_MASK_COLOR = _get('WIZARD_MASK_COLOR', '[0.125, 0.909, 0.972]')
    "Color to use in some wizards."

    # Notification
    SCIPION_NOTIFY = _get('SCIPION_NOTIFY', 'True')
    "If set to False, Scipion developers will know almost nothing about Scipion usage and will have less information to improve it."

    SCIPION_CWD = _get('SCIPION_CWD', os.path.abspath(os.getcwd()))
    "Directory when scipion was launched"

    SCIPION_GUI_REFRESH_IN_THREAD = _get('SCIPION_GUI_REFRESH_IN_THREAD', 'False')
    "True to refresh the runs graph with a thread. Unstable."

    SCIPION_GUI_REFRESH_INITIAL_WAIT = int(_get("SCIPION_GUI_REFRESH_INITIAL_WAIT", 5))
    "Seconds to wait after a manual refresh"

    SCIPION_GUI_CANCEL_AUTO_REFRESH = _get("SCIPION_GUI_CANCEL_AUTO_REFRESH","False")
    "Set it to True to cancel automatic refresh of the runs."

    # Cancel shutil fast copy. In GPFS, shutil.copy does fail when trying a fastcopy and does not fallback on the slow copy.
    SCIPION_CANCEL_FASTCOPY = _get('SCIPION_CANCEL_FASTCOPY', None)
    "Cancel fast copy done by shutil (copying files) when it fails. Has happened in GPFS environments."

    # Priority package list: This variable is used in the view protocols in
    # order to load first the plugins that contains the main protocols.conf
    # sections, so other plugins can define only their sections avoiding
    # duplicating all the sections in all plugins
    SCIPION_PRIORITY_PACKAGE_LIST = _get('SCIPION_PRIORITY_PACKAGE_LIST', None)

    SCIPION_STEPS_CHECK_SEC = int(_get('SCIPION_STEPS_CHECK_SEC', 5))
    "Number of seconds to wait before checking if new input is available in streamified protocols."

    SCIPION_UPDATE_SET_ATTEMPTS = int(_get('SCIPION_UPDATE_SET_ATTEMPTS', 3))
    "Number of attempts to modify the protocol output before failing. The default value is 3"

    SCIPION_UPDATE_SET_ATTEMPT_WAIT = int(_get('SCIPION_UPDATE_SET_ATTEMPT_WAIT', 2))
    "Time in seconds to wait until the next attempt when checking new outputs. The default value is 2 seconds"

    try:
        VIEWERS = ast.literal_eval(_get('VIEWERS', "{}"))
    except Exception as e:
        VIEWERS = {}
        print("ERROR loading preferred viewers, VIEWERS variable will be ignored")
        print(e)

    SCIPION_DOMAIN = _get(SCIPION_DOMAIN, None)
    SCIPION_TESTS_CMD = _get(SCIPION_TESTS_CMD, getTestsScript())

    # ---- Getters ---- #
    # Getters are alternatives to offer a variable, but preventing it to be stored in the config
    @classmethod
    def getLibFolder(cls):
        """
        :return: Folder where libraries must be placed in case a binding needs them
        """
        lib = cls._join(cls.SCIPION_SOFTWARE, 'lib')
        os.makedirs(lib, exist_ok=True)
        return lib

    @classmethod
    def getBindingsFolder(cls):
        """
        Folder where bindings must be placed. This folder is added to sys.path at launching time.
        If the binding depends on a dynamic libraries, those must be placed at cls.getLibFolder()
        :return:   The bindings folder
        """
        bindings = cls._join(cls.SCIPION_SOFTWARE, 'bindings')
        os.makedirs(bindings, exist_ok=True)
        return bindings

    @classmethod
    def getLogsFolder(cls):
        """
        Folder where scipion logs must be placed. The folder is created         
        """
        logsFolder = cls.SCIPION_LOGS
        os.makedirs(logsFolder, exist_ok=True)
        return logsFolder

    @classmethod
    def getVars(cls):
        """ Return a dictionary with all variables defined
        in this Config.
        """
        configVars = dict()
        # For each variable, also in base classes
        for baseCls in inspect.getmro(cls):
            for name, value in vars(baseCls).items():
                # Skip methods and internal attributes starting with __
                # (e.g __doc__, __module__, etc)
                if (isinstance(value, str) or isinstance(value, int)) and not name.startswith('__'):
                    configVars[name] = str(value)
        return configVars

    @classmethod
    def printVars(cls):
        """ Print the variables dict, mostly for debugging. """
        from .utils import prettyDict
        prettyDict(cls.getVars())

    @classmethod
    def getDomain(cls):
        """ Import domain module from path or name defined in SCIPION_DOMAIN.
        """
        value = cls.SCIPION_DOMAIN

        if not value:
            return None

        if os.path.isdir(value):
            dirname, value = os.path.split(value)
            sys.path.append(dirname)

        return importlib.import_module(value).Domain

    @classmethod
    def setDomain(cls, moduleOrNameOrPath):
        if isinstance(moduleOrNameOrPath, types.ModuleType):
            value = os.path.abspath(moduleOrNameOrPath.__path__[0])
        else:
            value = moduleOrNameOrPath
        cls.SCIPION_DOMAIN = value
        os.environ[SCIPION_DOMAIN] = value

    @staticmethod
    def getPythonLibFolder():
        from sysconfig import get_paths
        return os.path.join(get_paths()['data'], "lib")

    @staticmethod
    def debugOn():
        """ Returns a True if debug mode (SCIPION_DEBUG variable) is active """
        from .utils import envVarOn
        return bool(envVarOn(SCIPION_DEBUG))

    @staticmethod
    def toggleDebug():
        debugOn = not Config.debugOn()
        os.environ[SCIPION_DEBUG] = str(debugOn)
        os.environ[SCIPION_DEBUG_NOCLEAN] = str(debugOn)
        os.environ[SCIPION_LOG_LEVEL] = "INFO" if not debugOn else "DEBUG"

    @staticmethod
    def debugSQLOn():
        from .utils import envVarOn
        return bool(envVarOn(SCIPION_DEBUG_SQLITE))

    @staticmethod
    def toggleDebugSQL():
        newValue = not Config.debugSQLOn()
        os.environ[SCIPION_DEBUG_SQLITE] = str(newValue)

    @classmethod
    def refreshInThreads(cls):
        from .utils import strToBoolean
        return strToBoolean(cls.SCIPION_GUI_REFRESH_IN_THREAD)

    @classmethod
    def getExternalJsonTemplates(cls):
        return os.path.dirname(cls.SCIPION_CONFIG)

    @classmethod
    def getWizardMaskColor(cls):
        return json.loads(cls.WIZARD_MASK_COLOR)

    @classmethod
    def getPriorityPackageList(cls):
        if cls.SCIPION_PRIORITY_PACKAGE_LIST is not None:
            return cls.SCIPION_PRIORITY_PACKAGE_LIST.split(" ")
        else:
            return []

    @classmethod
    def getStepsCheckSeconds(cls):
        return cls.SCIPION_STEPS_CHECK_SEC

    @classmethod
    def getUpdateSetAttempts(cls):
        return cls.SCIPION_UPDATE_SET_ATTEMPTS

    @classmethod
    def getUpdateSetAttemptsWait(cls):
        return cls.SCIPION_UPDATE_SET_ATTEMPT_WAIT

    @classmethod
    def colorsInTerminal(cls):
        """ Returns true if colors are allowed. Based on NO_COLOR variable. Undefined or '' colors are enabled"""
        return cls.NO_COLOR == ''


# Add bindings folder to sys.path
sys.path.append(Config.getBindingsFolder())

# Cancel fast copy
if Config.SCIPION_CANCEL_FASTCOPY:
    shutil._USE_CP_SENDFILE = False
